
OldFileStream subclass: #BufferedFileStream
  instanceVariableNames: 
    'buffer bufferSize bufferIndex bufferStart bufferLimit written '
  classVariableNames: ''
  poolDictionaries: 
    'CharacterConstants ' !

! BufferedFileStream class methods !

on: afile bufferSize: anInteger
        "Answer a new instance of the
         receiver on anArrayedCollection."
    ^(self new)
        on: afile;
        setLimits: anInteger! !


! BufferedFileStream methods !

close
        "Close the file associated with the receiver stream
         after writing all the data to the file."
    written at:bufferIndex put:writtenOn.
    self readLimit.
    self writeBuffer.
    file close!

flush
        "Guarantee that any writes to the receiver stream
         are physically recorded on disk."
    written at:bufferIndex put:writtenOn.
    self
        setToEnd;
        readLimit;
        writeBuffer.
    file flush!

position: anInteger
        "Set the receiver stream position to anInteger."
    self readLimit.
    ((anInteger + 1) < pageStart
        or: [(pageStart + File pageSize) <= (anInteger + 1)])
            ifTrue: [
                ((anInteger + 1) < bufferStart
                    or: [(bufferStart + (bufferSize * File pageSize)) <= (anInteger + 1)])
                        ifTrue: [
                            self writeBuffer.
                            bufferStart := ((anInteger // (bufferSize * File pageSize))
                                * (bufferSize * File pageSize)) + 1.
                            self readBuffer.
                            pageStart := ((anInteger // File pageSize)
                                * File pageSize) + 1.
                            self update]
                        ifFalse: [
                            written at: bufferIndex put: writtenOn.
                            pageStart := ((anInteger // File pageSize)
                                * File pageSize) + 1.
                            self update]].
    position := anInteger - pageStart + 1.
    self updateLastByte!

readBuffer
        "Private - Read into buffer."
    | partialRead bytesRead pageSize loc index|
    pageSize := File pageSize.
    partialRead := false.
    loc := bufferStart.
    index := 1.
    [(index > bufferSize) or:[partialRead]]
        whileFalse: [
             bytesRead := file
                readBuffer: (buffer at: index)
                atPosition: loc.
             index := index + 1.
             loc := loc + pageSize.
             bytesRead = pageSize ifFalse:[partialRead := true]].
    partialRead
        ifTrue:[bufferLimit := (index - 2) * pageSize + bytesRead]
        ifFalse:[bufferLimit := (bufferSize * pageSize)]!

readLimit
        "Private - Answer the size of the receiver file."
    readLimit := readLimit max: position.
    bufferLimit := bufferLimit max: ((bufferIndex - 1) * File pageSize + position).
    ^self updateLastByte!

setLimits: anInteger
        "Private - Initialize the stream size and position."
    | partialRead bytesRead pageSize |
    file := collection.
    buffer := Array new: anInteger.
    written := Array new: anInteger.
    bufferSize := anInteger.
    pageSize := File pageSize.
    1 to:bufferSize do:[:index|
        buffer at:index put:(String new:pageSize).
        written at:index put:false].
    writeLimit := pageSize.
    partialRead := false.
    position := 0.
    bufferIndex := 1.
    [(bufferIndex > bufferSize) or:[partialRead]]
        whileFalse: [
             bytesRead := file fileId
                readInto: (buffer at: bufferIndex)
                atPosition: position.
             bufferIndex := bufferIndex + 1.
             position := position + pageSize.
             bytesRead = pageSize ifFalse:[partialRead := true]].
    partialRead
        ifTrue:[bufferLimit := (bufferIndex - 2) * pageSize + bytesRead]
        ifFalse:[bufferLimit := pageSize * bufferSize].
    bufferLimit < pageSize
        ifTrue:[readLimit := bufferLimit]
        ifFalse:[readLimit := pageSize].
    bufferIndex := 1.
    collection := buffer at:bufferIndex.
    position := 0.
    pageStart := 1.
    bufferStart := 1.
    writtenOn := false.
    lastByte := file size.
    lineDelimiter := Cr.
    2 to: readLimit do: [ :i |
        (collection at: i) == Lf
            ifTrue: [
                (collection at: i - 1) == Cr
                    ifFalse: [lineDelimiter := Lf].
                ^self]]!

update
    | pageSize |
    pageSize := File pageSize.
    bufferIndex := (pageStart - bufferStart) // pageSize + 1.
    collection := buffer at:bufferIndex.
    writtenOn := written at:bufferIndex.
    (bufferLimit = (bufferSize * pageSize))
        ifTrue:[readLimit := pageSize]
        ifFalse:[
            (bufferLimit - (bufferLimit \\ pageSize)) > (pageStart - bufferStart)
                ifTrue:[readLimit := pageSize]
                ifFalse:[(bufferLimit - (bufferLimit \\ pageSize)) < (pageStart - bufferStart)
                    ifTrue:[self error:'position outside of stream']
                    ifFalse:[readLimit := bufferLimit \\ pageSize]]]!

writeBuffer
    | index saved endIndex loc pageSize|
    pageSize := File pageSize.
    saved := readLimit.
    written at:bufferIndex put:writtenOn.
    index := 0.
    loc := bufferStart.
    endIndex := (bufferLimit - (bufferLimit \\ pageSize)) // pageSize + 1.
    [(index := index + 1) < endIndex]
            whileTrue:[(written at:index)
                            ifTrue:[file writeBuffer:(buffer at:index)
                                ofSize:pageSize
                                atPosition:loc].
                        loc := loc + pageSize].
    readLimit := bufferLimit \\ pageSize.
    readLimit = 0
        ifFalse:[(written at: endIndex)
            ifTrue:[file writeBuffer:(buffer at:endIndex)
                ofSize:readLimit
                atPosition:loc]].
    index := 0.
    [(index := index + 1) <= bufferSize]
        whileTrue:[written at:index put:false].
    collection := buffer at:bufferIndex.
    readLimit := saved.
    writtenOn := false! !
