
"This file contains the Cloner classes.

 The Cloner does a complete memory walk of all of the objects
 in Smalltalk/V.  It will purge all unused Symbols.  In addition
 it reassigns all of the object pointers to the low end of the
 object pointer space.

 Note: When you clone, the IMAGE FILE WILL BE REWRITTEN.
 Therefore, you should back up your image file before cloning.

 Cloning takes a long time (hours). To let you know that
 everything is alright, the Cloner writes a dot on the screen
 for every 5 objects cloned.

 To clone Smalltalk, make sure that this file is in your
 Disk directory, (the same directory that the image file
 is in).  Then evaluate the following expression in
 any TextPane, (e.g. the System Transcript):

    (Disk file: 'cloner.st') fileIn; close.

 When cloning is finished Smalltalk will exit to DOS.
 You should make a back up of your image file"


Smalltalk
    removeKey: #Cloner ifAbsent: [];
    removeKey: #CloneSet ifAbsent: [];
    removeKey: #CloneDictionary ifAbsent: [];
    removeKey: #CloneIdentityDictionary ifAbsent: [];
    at: #Clone put: nil !

Set variableSubclass: #CloneSet
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: '' !

!CloneSet class methods ! !

!CloneSet methods !

add: anObject
    (Clone checkClassOf: anObject)
        ifTrue: [^super add: anObject].
    ^nil!

findElementIndex: anObject
    | index indexedObject lastIndex |
    lastIndex := self basicSize.
    index := (Clone hashFor: anObject) \\ lastIndex + 1.
    [(indexedObject := self basicAt: index) = anObject]
        whileFalse: [
            (indexedObject == nil)
                ifTrue: [^index].
            (index := index + 1) > lastIndex
                ifTrue: [ "index wraparound"
                    index := 1]].
    ^index! !

Dictionary variableSubclass: #CloneDictionary
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: '' !

!CloneDictionary class methods ! !

!CloneDictionary methods !

add: anAssoc
    (Clone checkClassOf: anAssoc key)
        ifTrue: [
            (Clone checkClassOf: anAssoc value)
                ifTrue: [^super add: anAssoc]
                ifFalse: [
                    ^super add: (Association key: anAssoc key value: nil)]]
        ifFalse: [^nil]!

findKeyIndex: key
    | index indexedAssoc lastIndex |
    lastIndex := self basicSize.
    index := (Clone hashFor: key) \\ lastIndex + 1.
    [((indexedAssoc := self basicAt: index) == nil)
        or: [key = indexedAssoc key]]
            whileFalse: [
                (index := index + 1) > lastIndex
                    ifTrue: [ "index wraparound"
                        index := 1]].
    ^index! !

IdentityDictionary variableSubclass: #CloneIdentityDictionary
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: '' !

!CloneIdentityDictionary class methods ! !

!CloneIdentityDictionary methods !

add: anAssoc
    (Clone checkClassOf: anAssoc key)
        ifTrue: [
            (Clone checkClassOf: anAssoc value)
                ifTrue: [^super add: anAssoc]
                ifFalse: [
                    ^super add: (Association key: anAssoc key value: nil)]]
        ifFalse: [^nil]!

findKeyIndex: key
    | index lastIndex |
    lastIndex := self basicSize.
    index := ((Clone hashFor: key) bitAnd: lastIndex - 1) + 1.
    [key == (self basicAt: index)]
        whileFalse: [
            (self basicAt: index) == nil
                ifTrue: [^index].
            (index := index + 1) > lastIndex
                ifTrue: [ "index wraparound"
                    index := 1]].
    ^index! !

Object subclass: #Cloner
  instanceVariableNames:
    'clonePointer clones objectsSeen clonesToBeDumped classesToSkip
    symbolSetPosition cloneClasses'
  classVariableNames: ''
  poolDictionaries: '' !

!Cloner class methods ! !

!Cloner methods !

addSkipClasses: collection
    "Add the classes for each symbol
     in collection to cloneClassesToSkip"

    | class |
    collection do: [ :symbol |
        class := Smalltalk at: symbol.
        classesToSkip
            add: class;
            add: class class.
        objectsSeen
            add: class class]!

checkClassOf: anObject
    (anObject isKindOf: Behavior)
        ifTrue: [^cloneClasses includes: anObject].
    ^cloneClasses includes: anObject class!

cloneClasses
    ^cloneClasses!

cloneFixedPointers
    self
        cloneOopFor: false;
        cloneOopFor: 'schedAssocPtr';
        cloneOopFor: Character;
        cloneOopFor: nil;
        cloneOopFor: Smalltalk;
        cloneOopFor: Symbol symbolTable;
        cloneOopFor: String;
        cloneOopFor: UndefinedObject;
        cloneOopFor: #value;
        cloneOopFor: #value:;
        cloneOopFor: Symbol;
        cloneOopFor: Context;
        cloneOopFor: Array;
        cloneOopFor: FileControlBlock;
        cloneOopFor: Point;
        cloneOopFor: Message;
        cloneOopFor: true;
        cloneOopFor: CompiledMethod;
        cloneOopFor: Process;
        cloneOopFor: #mustBeBoolean;
        cloneOopFor: SmallInteger;
        cloneOopFor: #x;
        cloneOopFor: #y;
        cloneOopFor: #stackOverflow;
        cloneOopFor: Form;
        cloneOopFor: DisplayScreen;
        cloneOopFor: DisplayMedium;
        cloneOopFor: InfiniteForm;
        cloneOopFor: DisplayString;
        cloneOopFor: #doesNotUnderstand:;
        cloneOopFor: #cantReturn;
        cloneOopFor: #controlBreak;
        cloneOopFor: LargeNegativeInteger;
        cloneOopFor: #<;
        cloneOopFor: #>;
        cloneOopFor: #<=;
        cloneOopFor: #=;
        cloneOopFor: #+;
        cloneOopFor: #-;
        cloneOopFor: #*;
        cloneOopFor: #//;
        cloneOopFor: #\\;
        cloneOopFor: #bitAnd:;
        cloneOopFor: #bitShift:;
        cloneOopFor: #@;
        cloneOopFor: LargePositiveInteger;
        cloneOopFor: Compiler;
        cloneOopFor: #startUp.
    0 to: 255 do: [ :i | self cloneOopFor: i asCharacter]!

cloneOn: aStream

    | count clonesize dumpsize dumpObj |
    count := clonesize := dumpsize := 0.
    objectsSeen := classesToSkip := nil.
    [clonesToBeDumped isEmpty] whileFalse: [
        (clones size > clonesize) ifTrue: [
            clonesize := clones size.
            clonesize printString displayAt: 0 @ 21].
        (clonesToBeDumped size > dumpsize) ifTrue: [
            dumpsize := clonesToBeDumped size.
            dumpsize printString displayAt: 0 @ 22].
        ('clones size => ' , '     '
            replaceTo: 20
            with: (clones size printString))
        displayAt: 0 @ 23.
        ('clonesToBeDumped size => ' , '     '
            replaceTo: 30
            with: (clonesToBeDumped size printString))
        displayAt: 0 @ 24.
        (count := count + 1 \\ 5) == 0 ifTrue: [Terminal write: $.].
        dumpObj := clonesToBeDumped removeFirst.
        (dumpObj isKindOf: Behavior)
            ifTrue: [dumpObj printString , '     ' displayAt: 0 @ 20]
            ifFalse: [
                'a ' , (dumpObj class printString) , '    ' displayAt: 0 @ 20].
        dumpObj dumpOn: aStream for: self]!

cloneOopFor: anObject
    "Answer the cloned oop for anObject"

    anObject == nil
        ifTrue: [
            clonePointer = 7
                ifTrue: [
                    clonesToBeDumped addLast: anObject.
                    clonePointer := 9].
            ^7 ].
    anObject class == SmallInteger
        ifTrue: [^anObject * 2].
    (self checkClassOf: anObject)
        ifTrue: [
            ^clones at: anObject ifAbsent: [
                clones at: anObject put: clonePointer.
                clonePointer := clonePointer + 2.
                clonesToBeDumped addLast: anObject.
                ^clonePointer - 2]]
        ifFalse: [^7]!

cloneSystemOn: aStream
    "Clone the system on aStream"

    | newSymbolSet |
    self initializeClones.
    self skipClasses.
    aStream
        nextTwoBytesPut: 16r11. "Smalltalk/V R1.1"
    aStream
        nextTwoBytesPut: 0.     "Reserve the first two bytes into"
                                "which ClonerPointer will be placed"
                                "after the entire dump"
    objectsSeen do: [ :object |
        clones at: object put: 7 ].
    self cloneFixedPointers.
        "reserving 3000 free oops"
    clonePointer := clonePointer + 6000.
    clonesToBeDumped add: 6000.
    self cloneSelectors.
    self cloneOn: aStream.
    clonesToBeDumped := nil.
    aStream position: symbolSetPosition.
    newSymbolSet := SymbolSet new: Symbol symbolTable basicSize.
    clones keysDo: [ :each |
        each class == Symbol
            ifTrue: [newSymbolSet add: each]].
    self dump: newSymbolSet on: aStream.
    aStream
        position: 2;
        nextTwoBytesPut: clonePointer;
        flush.
    Symbol symbolTable do: [ :each |
        (newSymbolSet includes: each) ifFalse: [
            Transcript nextPutAll: each; cr]].
    clonePointer := clones := cloneClasses := nil!

dump: anObject on: aStream
    "anObject dumped as a stream of bytes. Called by SymbolSet."

    | index |
    aStream nextTwoBytesPut: (self cloneOopFor: anObject class).
    index := anObject class instSize + anObject basicSize.
    aStream nextTwoBytesPut: index + 1 * 2.
    1 to: index do: [ :i |
        aStream nextTwoBytesPut:
            (self cloneOopFor: (anObject instVarAt: i))]!

dump: anObject on: aStream for: aClass
    "Dumped anObject as a stream of bytes
     changing the class to aClass"

    | index |
    aStream nextTwoBytesPut: (self cloneOopFor: aClass).
    index := anObject class instSize + anObject basicSize.
    aStream nextTwoBytesPut: index + 1 * 2.
    1 to: index do: [ :i |
        aStream nextTwoBytesPut:
            (self cloneOopFor: (anObject instVarAt: i))]!

hashFor: anObject
    "Hash using cloned object pointer if appropriate"

    | hash |
    (hash := anObject hash) = anObject basicHash
        ifTrue:  [^(self cloneOopFor: anObject) // 2 bitAnd: 16r3FFF]
        ifFalse: [
            ((anObject isKindOf: SequenceableCollection)
                and: [anObject class ~~ String])
                    ifTrue: [
                        ^(anObject size = 0
                          or: [anObject == (anObject at: 1)])
                            ifTrue: [7]
                            ifFalse: [self hashFor: (anObject at: 1)]]
                    ifFalse: [^hash]]!

initializeClones
    "Initialize the cloning database"

    clonePointer := 1.
    clones := IdentityDictionary newSize: 8192.
    objectsSeen := Set new.
    clonesToBeDumped := OrderedCollection new.
    classesToSkip := Set new.
    cloneClasses := Object allSubclasses.
    cloneClasses add: Object.
    cloneClasses addAll: Object class allSubclasses.
    cloneClasses add: Object class.
    cloneClasses := cloneClasses asSet.
    self addSkipClasses:
        #(Cloner CloneSet CloneDictionary CloneIdentityDictionary).
    classesToSkip do: [ :class | cloneClasses remove: class ifAbsent: []]!

newCloneDictionary: anInteger
    ^CloneDictionary new: anInteger!

newCloneIdentityDictionary: anInteger
    ^CloneIdentityDictionary newSize: anInteger!

newCloneSet
    ^CloneSet new!

skipClasses
    classesToSkip do: [ :each |
        each superclass subclasses:
            (each superclass subclasses asOrderedCollection
                remove: each;
                yourself)]!

symbolSetPosition: anInteger
    symbolSetPosition := anInteger!

cloneSelectors
    | selectors |
    selectors := Set new: 1000.
    cloneClasses do: [ :class |
        selectors addAll: class methodDictionary keys].
    selectors := selectors asOrderedCollection.
    [selectors size > 0]
        whileTrue: [
            self cloneOopFor:
                selectors removeFirst.
            selectors size > 0
                ifTrue: [
                    self cloneOopFor:
                        selectors removeLast]]! !

"the following expression does the actual cloning"
| file |
Smalltalk unusedMemory.
file := DiskB newFile: 'image' bufferSize: 10.
(Clone := Cloner new) cloneSystemOn: file.
file close.
Smalltalk exit: true !

