
Inspector subclass: #Debugger
  instanceVariableNames: 
    'process walkback selectors method stream walkbackIndex walkbackPane temps source logger label resumable '
  classVariableNames: ''
  poolDictionaries: '' !

!Debugger class methods ! !


!Debugger methods !

accept: aString from: aDispatcher
        "Private - Compile and evaluate the code in aString.
         Replace the selected temporary variable with
         the result of the evaluation.  Answer true if
         there is no compilation error, else notify
         aDispatcher as to the error and answer false."
    | result context |
    result := Compiler
        evaluate: aString
        in: object class
        to: object
        notifying: aDispatcher
        ifFail: [^false].
    Smalltalk logEvaluate: aString.
    instIndex = 1
        ifTrue: [^true].
    (context := process
        frameAt: walkbackIndex offset: 5) isNil
        ifTrue: [ "addressing stack"
            process frameAt: walkbackIndex
                offset: 6 + instList size
                    - instIndex
                put: result]
        ifFalse: [ "addressing context"
            context := context homeContext.
            context
                at:  context size - instIndex + 2
                put: result].
    self changed: #instance.
    ^true!

atEnd
        "Private - Test for end of source stream while
         skipping over white space, comments and strings."
    [(stream peekFor: $")
        ifTrue: [stream skipTo: $"].
    (stream peekFor: $')
        ifTrue: [stream skipTo: $'].
    stream atEnd]
        whileFalse: [
            stream peek isSeparator
                ifTrue: [stream next]
                ifFalse: [^false]].
    ^true!

cr
        "Terminate current walkback line."
    walkback add: stream contents.
    stream reset!

debug
        "Private - Close the walkback window and
         open a debugger window on the process."
    | aTopPane |
    logger topDispatcher closeWindow.
    Scheduler displayAll.
    aTopPane := TopPane new
        label: label;
        yourself.
    aTopPane addSubpane:
        (walkbackPane := ListPane new
            model: self;
            name: #walkback;
            change: #walkback:;
            returnIndex: true;
            menu: #menu;
            framingRatio: (0@0
                extent: 8/15 @ (2/5));
            yourself).
    aTopPane addSubpane:
        (ListPane new
            menu: #inspectMenu;
            model: self;
            name: #tempList;
            change: #selectInstance:;
            returnIndex: true;
            framingRatio: (8/15@0
                extent: 2/15@(2/5))).
    aTopPane addSubpane: (instPane := TextPane new
        model: self;
        name: #instance;
        change: #accept:from:;
        framingRatio: (2/3@0
            extent: 1/3@(2/5)));
        yourself.
    aTopPane addSubpane:
        (TextPane new
            model: self;
            name: #method;
            change: #take:from:;
            framingRatio: (0@(2/5)
                extent: 1@(3/5))).
   logger := aTopPane dispatcher.
   logger open scheduleWindow!

implementors
        "Private - Pop-up a window with the implementors
         of the method for the selected line."
    walkbackIndex isNil
        ifFalse: [
            Smalltalk implementorsOf:
                (selectors at: walkbackIndex) key]!

inspectSelection
        "Private - Open an Inspector on the
         selected temporary variable."
    (walkbackIndex isNil or: [instIndex isNil])
        ifTrue: [^self].
    self tempValue inspect!

instance
        "Private - Answer the ASCII representation
         of the selected temporary variable."
    (walkbackIndex isNil or: [instIndex isNil])
        ifTrue: [^String new].
    ^self tempValue printString!

label: aString
        "Pop-up a walkback window with the label equal
         to aString.  Display the stacked message
         sends for the current process in the window."
    ^self walkBackFor: Process copyStack label: aString!

menu
        "Private - Answer the walkback pane
         menu for the debugger window."
    ^Menu
        labels: 'resume\restart\senders\implementors' withCrs
        lines: #(2)
        selectors: #(resume restart senders implementors)!

method
        "Private - Answer a string containing the source
         code for the selected method in the walkback."
    walkbackIndex isNil
        ifTrue: [
            method := nil.
            ^String new].
    ^source!

nextPut: aCharacter
        "Append aCharacter to the walkback line."
    stream nextPut: aCharacter!

nextPutAll: aString
        "Append aString to the walkback line."
    stream nextPutAll: aString!

restart
        "Private - Restart the process
         at the selected method."
    walkbackIndex isNil
        ifTrue: [^self].
    walkbackIndex = (process homeFrameOf: walkbackIndex)
        ifFalse: [
            ^Menu message: 'Can''t restart at block.'].
    ((process methodAt: walkbackIndex) at: 1) = 0
        ifFalse: [
            ^Menu message: 'Can''t restart primitive method.'].
    logger topDispatcher closeWindow.
    Scheduler displayAll.
    process restartAt: walkbackIndex!

resumable: aBoolean
        "Mark the resumability of the
         debugged process to aBoolean."
    resumable := aBoolean!

resume
        "Private - Resume the process if resumable."
    resumable
        ifTrue: [
            logger topDispatcher closeWindow.
            Scheduler displayAll.
            process resume]
        ifFalse: [
            Menu message:
                'Process is not resumable, but may be restarted at a selected method.']!

scanArgs: anInteger
        "Private - Scan arguments in the source code
         of the selected method to add anInteger number
         of names to the temps collection."
    | aString |
    anInteger = 0
        ifTrue: [self skipName].
    anInteger timesRepeat: [
        self skipName.
        self skipPunctuation.
        (aString := stream nextWord) isNil
            ifTrue: [^self]
            ifFalse: [temps add: aString]]!

scanBlocks: numberArgsAndTemps
        "Private - Scan the source code of the selected
         method to parse the block argument names and add
         them to the temps collection."
    | aString |
    [numberArgsAndTemps = temps size or: [self atEnd]]
        whileFalse: [
            [self atEnd or: [stream next = $[]]
                whileFalse: [].
            [self atEnd or: [stream next ~= $:]]
                whileFalse: [
                    (aString := stream nextWord) isNil
                        ifFalse: [temps add: aString]]]!

scanTemps: anInteger
        "Private - Scan the source code of the
         selected method to parse the temporary
         variable declarations and add them to
         the temps collection."
    | aString |
    self atEnd
        ifTrue: [^self].
    (stream peekFor: $|)
        ifTrue: [
            anInteger timesRepeat: [
                self atEnd.
                ((stream peekFor: $|) or:
                [(aString := stream nextWord) isNil])
                    ifTrue: [^self].
                temps add: aString]]!

senders
        "Private - Pop-up a window with
         the senders of the selected method."
    walkbackIndex isNil
        ifFalse: [
            Smalltalk sendersOf:
                (selectors at: walkbackIndex) key]!

skipName
        "Private - Skip over the next name in the
         source code for the selected method."
    [self atEnd not and: [stream peek isAlphaNumeric]]
        whileTrue: [stream next]!

skipPunctuation
        "Private - Skip to the next letter in the
         source code for the selected method."
    | end |
    [(end := self atEnd) or: [stream next isLetter]]
        whileFalse: [].
    end ifFalse: [stream skip: -1]!

take: aString from: aDispatcher
        "Private - Accept aString as an updated
         method and compile it.  Notify aDispatcher
         if the compiler detects errors."
    | result aClass oldMethod |
    walkbackIndex isNil
        ifTrue: [^true].
    CursorManager execute change.
    oldMethod := process methodAt: walkbackIndex.
    aClass := (selectors at: walkbackIndex) value.
    result := aClass compile: aString
        notifying: aDispatcher.
    result isNil
        ifTrue: [^false].
    Smalltalk
        logSource: aString
        forSelector: result key
        inClass: aClass.
    (selectors at: walkbackIndex) key = result key
        ifFalse: [^true].
    process dropTo: oldMethod.
    process methodAt: 1
        put: (aClass compiledMethodAt: result key).
    resumable := false.
    self
        walkback;
        walkback: 1.
    walkbackPane restoreSelected: walkback first.
    CursorManager normal change.
    ^true!

tempList
        "Private - Scan the source code of the selected
         method to build an OrderedCollection of the
         method temporary variables and argument names
         in declaration order.  Assign the collection
         to instance variable temps.  Answer the contents
         of temps."
    | numberArgs numberTemps assoc aClass selector |
    walkbackIndex isNil
        ifTrue: [
            method := nil.
            ^String new].
    assoc := selectors at: walkbackIndex.
    selector := assoc key.
    aClass := assoc value.
    method := aClass compiledMethodAt: selector.
    source := aClass sourceCodeAt: selector.
    temps := OrderedCollection new.
    instIndex := nil.
    stream := ReadStream on: source.
    numberArgs := method at: 3.
    (numberTemps := method at: 2) > 100
        ifTrue: [numberTemps := 255 - numberTemps].
    self scanArgs: numberArgs.
    self scanTemps: numberTemps.
    self scanBlocks: numberArgs + numberTemps.
    temps size + 1 to: numberArgs do: [ :i |
        temps add: '.arg', i printString].
    temps size + 1 - numberArgs to:
        numberTemps do: [ :i |
            temps add: '.temp', i printString].
    temps addFirst: 'self'.
    ^instList := temps!

tempValue
        "Private - Answer the value of the
         selected temporary variable."
    | context |
    instIndex = 1
        ifTrue: [^object].
    ^(context := process
        frameAt: walkbackIndex offset: 5) isNil
        ifTrue: [ "addressing stack"
            process frameAt: walkbackIndex
                offset: 6 + instList size
                - instIndex]
        ifFalse: [ "addressing context"
            context homeContext isNil
                ifFalse: [
                    context := context homeContext].
            context at: context size - instIndex + 2]!

walkback
        "Private - Answer an OrderedCollection
         of walkback lines."
    | method aClass index size |
    walkback := OrderedCollection new.
    stream := WriteStream on: String new.
    process walkBackOn: self.
    size := walkback size.
    selectors := OrderedCollection new.
    index := 1.
    [index > size or: [(process frameAt: index offset: 1) = 0]]
        whileFalse: [
            method := process methodAt: index.
            aClass := (process receiverAt: index) class.
            selectors add: (aClass selectorFor: method).
            index := index + 1].
    ^walkback!

walkback: anInteger
        "Private - Select the walkback line
         at index position anInteger."
    walkbackIndex := anInteger.
    object := process receiverAt: anInteger.
    self
        changed: #tempList;
        changed: #instance;
        changed: #method!

walkBackFor: aProcess label: aString
        "Pop-up a walkback window with label equal
         to aString).  Display the stacked message
         sends for the receiver in the window."
    | recursing extent |
    process := aProcess.
    recursing := RecursionInError.
    RecursionInError
        ifTrue: [
            logger := Terminal.
            logger cr; nextPutAll: 'error: ',aString; cr]
        ifFalse: [
            RecursionInError := true.
            extent := ((Display width // 2 max:
                aString size * LabelFont width + 6) min:
                Display width)
                @ (Display height // 2).
            label := aString.
            logger := TextEditor
                windowLabeled: aString
                frame: (Display extent - extent // 2
                    extent: extent).
            logger pane
                menu: #walkBackMenu;
                model: self].
    aProcess walkBackOn: logger.
    RecursionInError := false.
    recursing
        ifTrue: [^Smalltalk exit: false]
        ifFalse:[
            logger
                positionAtBeginning;
                scheduleWindow]!

walkBackMenu
        "Private - Answer the walkback pane
         menu for the walkback window."
    ^Menu
        labels: 'resume\debug' withCrs
        lines: Array new
        selectors: #(resume debug)! !
