
Object subclass: #File
  instanceVariableNames: 
    'name fileID directory fileSize attribute timeAndDate '
  classVariableNames: ''
  poolDictionaries: '' !

! File class methods !

changeModeOf: aString to: attrString
    "Change the attributes of the file named aString
     to those of attrString.  Attributes are:
     $r - read only, $h - hidden, $s - system,
     $a - archive (see DOS manual)."
    | aByte |
    aByte := 0.
    (attrString includes: $r)
        ifTrue: [aByte := aByte + 1].
    (attrString includes: $h)
        ifTrue: [aByte := aByte + 2].
    (attrString includes: $s)
        ifTrue: [aByte := aByte + 4].
    (attrString includes: $a)
        ifTrue: [aByte := aByte + 32].
    self dosError: (Dos
        changeModeOf: aString asAsciiZ
        to: aByte)!

fileName: nString extension: eString
    "Answer a String which is a derived file name
     from nString and eString. Lower case
     vowels are dropped from the right of nString
     until it is less than or equal to 8 characters."
    | size aString |
    size := nString size.
    aString :=
        (nString reversed select: [ :char |
        (char isVowel and: [char isLowerCase])
            ifTrue: [(size := size - 1) < 8]
            ifFalse: [true]]) reversed, '       '.
    ^(aString copyFrom: 1 to: 8),
        (String with: $.),
        ((eString, '   ') copyFrom: 1 to: 3)!

open: pathName type: access
    "Answer an FileStream of the receiver opened
     on a file named aString in aDisk."
    ^(self pathName: pathName) open: access!

pathName: aString
   "Answer a unopened File with path name aString."
    ^self pathName: aString in: Disk!

remove: aString
    "Erase the file named aString."
    self dosError:
        (Dos asciizPrimitives: 16r4100 with: aString asAsciiZ)!

rename: oldString to: newString
    "Rename the file named oldString to newString.
     If drives are specified in the names, they
     must refer to the same one."
    self dosError: (Dos
        rename: oldString asAsciiZ to: newString asAsciiZ)!

splitPath: aString
        "Answer an array of three strings.
         Drive, directory and name"
    | index1 index2 drive dirString fileString |
    index1 := 1.
    index2 := 0.
    (aString at: 2) == $:
        ifTrue: [
            drive := Diskette checkDrive: (aString at: 1).
            index1 := 2]
        ifFalse: [
            drive := Disk drive]! !


! File methods !

open: type
    "<primitive: >"! !
