
Object subclass: #Inspector
  instanceVariableNames: 
    'object instList instIndex instPane '
  classVariableNames: ''
  poolDictionaries: '' !

!Inspector class methods ! !


!Inspector methods !

accept: aString from: aDispatcher
        "Private - Compile and evaluate the code in aString.
         Replace the selected instance variable with the result
         of the evaluation.  Answer true if there is no compilation
         error, else notify aDispatcher as to the error and answer
         false."
    | result |
    result := Compiler
        evaluate: aString
        in: object class
        to: object
        notifying: aDispatcher
        ifFail: [^false].
    Smalltalk logEvaluate: aString.
    instIndex = 1
        ifFalse: [
            object class isBytes
                ifTrue: [
                    object
                        at: (instIndex - 1)
                        put: result]
                ifFalse: [
                    object
                        instVarAt: (instIndex  - 1)
                        put: result]].
    self changed: #instance.
    ^true!

doIt: aBlock
        "Private - Evaluate the selected text
         in the context of the inspected object."
    | aString result selection |
    aString := instPane selectedString.
    result := Compiler
        evaluate: aString
        in: object class
        to: object
        notifying: instPane dispatcher
        ifFail: aBlock.
    Smalltalk logEvaluate: aString.
    selection := instPane selection.
    instPane
        hideSelection;
        selectAfter: selection corner;
        displayChanges;
        showSelection.
    ^result!

doItResult: aTextPane error: aBlock
        "Private - Evaluate the selected text in aTextPane
         in the context of the inspected object."
    ^Compiler
        evaluate: aTextPane selectedString
        in: object class
        to: object
        notifying: aTextPane dispatcher
        ifFail: aBlock!

inspectMenu
        "Private - Answer the menu to
         inspect an instance variable."
    ^Menu
        labels: 'inspect'
        lines: Array new
        selectors: #(inspectSelection)!

inspectSelection
        "Private - Open an Inspector on the
         selected instance variable."
    instIndex = 1
        ifTrue:  [object inspect]
        ifFalse: [
            object class isBytes
                ifTrue: [(object at: (instIndex - 1)) inspect]
                ifFalse: [
                    (object instVarAt: (instIndex - 1)) inspect]]!

instance
        "Private - Answer the ASCII representation
         of the selected instance variable."
    instIndex = 1
        ifTrue: [^object printString].
    object class isPointers
       ifTrue: [^(object instVarAt: instIndex - 1) printString]
       ifFalse: [^(object at: instIndex - 1) printString]!

instVarList
        "Private - Answer an OrderedCollection of
         instance variable names and/or numbers for
         the list pane."
    ^instList!

openOn: anObject
        "Open an inspector window on anObject.  Define
         the pane sizes and behavior, and shedule the
         window."
    | aTopPane |
    object := anObject.
    instPane := TextPane new
        model: self;
        name: #instance;
        change: #accept:from:;
        framingRatio: (1/3 @ 0
            extent: 2/3 @ 1).
    aTopPane := TopPane new.
    aTopPane
        label: 'Inspecting: ', object class printString;
        model: aTopPane dispatcher;
        menu: #workSpaceMenu;
        minimumSize: 80@80;
        yourself.
    aTopPane addSubpane:
        (ListPane new
            menu: #inspectMenu;
            model: self;
            name: #instVarList;
            change: #selectInstance:;
            returnIndex: true;
            framingRatio: (
                0@0 extent: 1/3 @ 1)).
    aTopPane addSubpane: instPane.
    CursorManager normal change.
    self setInstList.
    aTopPane dispatcher open scheduleWindow!

remove
        "Private - Remove the selected
         key from the dictionary."
    instIndex isNil
        ifFalse: [
            object removeKey: (instList at: instIndex).
            self
                changed: #instVarList with: #restore;
                changed: #instance]!

selectInstance: anInteger
        "Private - Select the instance variable at
         index position anInteger in the list."
    instIndex := anInteger.
    self changed: #instance!

setInstList
        "Private - Initialize an OrderedCollection of
         instance variable names and/or numbers for
         the list pane."
    instIndex := 1.
    instList := OrderedCollection new.
    instList add: 'self'.
    instList := instList,
        object class allInstVarNames .
    object class isVariable
        ifTrue: [
            1 to: object basicSize do: [ :index |
                instList add: index printString]]! !
