
Number subclass: #Integer
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: '' !

! Integer class methods ! !


! Integer methods !

* aNumber
        "Answer the result of multiplying
         the receiver by aNumber."
    <primitive: 29>
    aNumber isSwapped
        ifTrue: [^self * aNumber].
    ^aNumber * self!

+ aNumber
        "Answer the sum of the receiver and aNumber."
    <primitive: 21>
    aNumber isSwapped
        ifTrue: [^self + aNumber].
    ^aNumber + self!

- aNumber
        "Answer the difference between
         the receiver and aNumber."
    <primitive: 22>
    aNumber isSwapped
        ifTrue: [^self - aNumber].
    ^(aNumber - self) negated!

/ aNumber
        "Answer the result of dividing
         the receiver by aNumber."
    | numerator denominator gcd |
    aNumber class == ComplexNumber
        ifTrue: [^self asComplexNumber / aNumber].
    aNumber class == Float
        ifTrue: [^self asFloat / aNumber].
    numerator := self * aNumber denominator.
    (denominator := aNumber numerator) < 0
        ifTrue: [
            denominator := 0 - denominator.
            numerator := 0 - numerator].
    (gcd := numerator gcd: denominator) = denominator
        ifTrue: [^numerator // gcd]
        ifFalse: [
            ^Fraction
                numerator: numerator // gcd
                denominator: denominator // gcd]!

// aNumber
        "Answer the qotient of dividing the
         receiver by aNumber with truncation
         towards negative infinity."
    | answer quotient |
    <primitive: 32>
    aNumber isSwapped
        ifTrue: [^self // aNumber].
    aNumber = 0 ifTrue: [^self zeroDivisor].
    quotient := self / aNumber.
    (answer := quotient truncated) < 0
        ifTrue: [
            answer = quotient
                ifFalse: [^answer - 1]].
    ^answer!

< aNumber
        "Answer true if the receiver is less
         than aNumber, else answer false."
    <primitive: 23>
    aNumber isSwapped
        ifTrue: [^self < aNumber].
    ^aNumber > self!

<= aNumber
        "Answer true if the receiver is less than
         or equal to aNumber, else answer false."
    <primitive: 25>
    aNumber isSwapped
        ifTrue: [^self <= aNumber].
    ^aNumber >= self!

= aNumber
        "Answer true if the receiver is equal
         to aNumber, else answer false."
    <primitive: 27>
    aNumber isSwapped
        ifTrue: [^self = aNumber].
    ^aNumber = self!

> aNumber
        "Answer true if the receiver is greater
         than aNumber, else answer false."
    <primitive: 24>
    aNumber isSwapped
        ifTrue: [^self > aNumber].
    ^aNumber < self!

>= aNumber
        "Answer true if the receiver is greater than
         or equal to aNumber, else answer false."
    <primitive: 26>
    aNumber isSwapped
        ifTrue: [^self >= aNumber].
    ^aNumber <= self!

\\ aNumber
        "Answer the integer remainder after dividing
         the receiver by aNumber with truncation
         towards negative infinity."
    <primitive: 31>
    aNumber isSwapped
        ifTrue: [^self \\ aNumber].
    aNumber = 0 ifTrue: [^self zeroDivisor].
    ^(self - (self // aNumber * aNumber)) truncated!

asCharacter
        "Answer the character whose ASCII encoding
         matches the value of the receiver."
    ^Character value: self!

asFloat
        "Answer the floating point
         representation of the receiver."
    ^Float fromInteger: self!

basicHash
        "Answer the hash value of the receiver
         which is the magnitude of the receiver."
    ^self abs!

bitAnd: anInteger
        "Answer an Integer representing the receiver
         bits ANDed with the argument anInteger."
    <primitive: 34>
    anInteger isSwapped
        ifTrue: [^self bitAnd: anInteger].
    ^self primitiveFailed!

bitAt: anInteger
        "Answer 0 if the bit at index position anInteger
         in the receiver is 0, else answer 1."
    1 > anInteger
        ifTrue: [
            self error: 'bitAt: index must be greater than zero'].
    ^1 bitAnd: (self bitShift: 1 - anInteger)!

bitInvert
        "Answer an integer whose bit values are
         the inverse of the bit values of the receiver."
    ^self bitXor: -1!

bitOr: anInteger
        "Answer an Integer representing the receiver
         bits ORed with the argument anInteger."
    <primitive: 35>
    anInteger isSwapped
        ifTrue: [^self bitOr: anInteger].
    ^self primitiveFailed!

bitShift: anInteger
        "Answer an integer which is the receiver bits
         shifted left anInteger number of bit positions
         if anInteger is positive, or bits shifted
         right for anInteger negated number of bit positions
         if anInteger is negative."
    <primitive: 37>
    anInteger isSwapped
        ifTrue: [^self bitShift: anInteger].
    ^self primitiveFailed!

bitXor: anInteger
        "Answer the receiver bit XORed
         with the argument anInteger."
    <primitive: 36>
    anInteger isSwapped
        ifTrue: [^self bitXor: anInteger].
    ^self primitiveFailed!

caseDo:alternativeBlocks otherwise:aBlock
        "Evaluate the block indexed by self."
    (self >= 1 and:[self <= alternativeBlocks size])
        ifTrue:[^(alternativeBlocks at:self) value]
        ifFalse:[^aBlock value]!

deepCopy
        "Answer a copy of the receiver with shallow
         copies of each instance variable.  Because
         integers are immutable (cannot instantiate
         a copy), answer the receiver."
    ^self!

dumpOn: aStream for: aCloner
        "Dump the receiver as a series of bytes to
         aStream for aCloner."
    | index |
    aStream nextTwoBytesPut: (aCloner cloneOopFor: self class).
    index := self basicSize.
    aStream nextTwoBytesPut: index + 1 * 2.
    1 to: index do: [ :i |
        aStream nextTwoBytesPut: (self at: i)]!

factorial
        "Answer the factorial of the receiver."
    self > 1
        ifTrue: [^(self - 1) factorial * self].
    self < 0
        ifTrue: [^self error: 'negative factorial'].
    ^1!

gcd: anInteger
        "Answer the greatest common divisor
         between the receiver and anInteger."
    | u v r |
    u := self abs.
    v := anInteger abs.
    [v = 0]
        whileFalse: [
            r := u \\ v.
            u := v.
            v := r].
    ^u!

hash
        "Answer the hash value of the receiver
         which is the absolute value of the receiver."
    ^self abs!

lcm: anInteger
        "Answer the least common multiple
         between the receiver and anInteger."
    | gcd |
    (gcd := self gcd: anInteger) = 0
        ifFalse: [^(self // gcd * anInteger) abs].
    ^1!

negated
        "Answer the receiver subtracted from zero."
    ^0 - self!

printOn: aStream
        "Append the ASCII representation (radix 10)
         of the receiver to aStream."
    | d divisor digits quotient remainder |
    self < 0
        ifTrue: [
            aStream nextPut: $-.
            0 - self printOn: aStream.
            ^self].
    digits := 4.
    divisor := 10000.
    [(d := divisor * divisor) < self]
        whileTrue: [divisor := d. digits := digits + digits].
    quotient := self // divisor.
    quotient printOn: aStream.
    remainder := (self - (quotient * divisor)) printString.
    aStream
        next: digits - remainder size put: $0;
        nextPutAll: remainder!

printOn: aStream base: anInteger
        "Append the ASCII representation of
         the receiver with radix b to aStream."
    | aString tempStream n |
    anInteger = 10
        ifFalse: ["print radix"
            anInteger printOn: aStream base: 10.
            aStream nextPut: $r].
    (n := self) < 0
        ifTrue: [
            aStream nextPut: $-.
            n := self negated].
    aString := String new: 16.
    tempStream := WriteStream on: aString.
    [anInteger <= n]
        whileTrue: [
            tempStream nextPut: (Character digitValue: n \\ anInteger).
            n := n // anInteger].
    aStream nextPut: (Character digitValue: n).
    n := tempStream position.
    [n = 0]
        whileFalse: [
            aStream nextPut: (aString at: n).
            n := n - 1]!

printPaddedTo: anInteger
        "Answer the string containing the ASCII representation
         of the receiver.  If the string is less than anInteger
         number of characters, the string is padded on the
         left so that the entire string is of size anInteger."
    | aStream padding digits|
    aStream := WriteStream on: (String new: 10).
    self printOn: aStream base: 10.
    digits := aStream contents.
    padding := anInteger - digits size.
    padding > 0
        ifTrue: [
            padding := ((String new: padding)
                atAllPut: $ ;
                yourself)]
        ifFalse:[padding := ''].
    ^padding, digits!

quo: aNumber
        "Answer the integer quotient of the receiver
         divided by aNumber with truncation toward zero."
    <primitive: 33>
    aNumber isSwapped
        ifTrue: [^self quo: aNumber].
    aNumber = 0 ifTrue: [^self zeroDivisor].
    ^(self / aNumber) truncated!

radix: anInteger
        "Answer a string which is the ASCII representation
         of the receiver with radix anInteger."
    | aStream |
    aStream := WriteStream on: (String new: 16).
    self printOn: aStream base: anInteger.
    ^aStream contents!

reciprocal
        "Answer one divided by the receiver."
    ^1 / self!

rem: aNumber
        "Answer the integer remainder after dividing
         the receiver by aNumber with truncation
         towards zero."
    <primitive: 20>
    aNumber isSwapped
        ifTrue: [^self rem: aNumber].
    aNumber = 0 ifTrue: [^self zeroDivisor].
    ^self - (self / aNumber * aNumber) truncated!

rounded
        "Answer the receiver."
    ^self!

shallowCopy
        "Answer a copy of the receiver which shares
         the receiver instance variables.  Because
         integers are immutable (cannot instantiate
         a copy), answer the receiver."
    ^self!

timesRepeat: aBlock
        "Evaluate the block aBlock n number
         of times, where n is the receiver."
    | anInteger |
    anInteger := self.
    [anInteger > 0] whileTrue: [
        anInteger := anInteger - 1.
        aBlock value]!

truncated
        "Answer the receiver."
    ^self!

zeroDivisor
        "Report an error to the effect that
         division by zero was attempted."
    self error: 'divisor is zero'!

~= aNumber
        "Answer true if the receiver is not
         equal to aNumber, else answer false."
    <primitive: 28>
    aNumber isSwapped
        ifTrue: [^self ~= aNumber].
    ^true! !
