
Object subclass: #OldDirectory
  instanceVariableNames: 
    'drive pathName volumeLabel '
  classVariableNames: ''
  poolDictionaries: '' !

! OldDirectory class methods !

current
    "Answer a Directory representing
     the current DOS directory."
    | aString index |
    aString := String new: 65.
    self dosError:
        (Dos getCurrentDirectory: Disk drive into: aString ).
    index := 1.
    aString := aString asLowerCase.
    [(aString at: index) asciiValue > 0]
        whileTrue: [index := index + 1].
    ^OldDirectory new
        drive: (Disk drive);
        pathName: (aString copyFrom: 1 to: index - 1)!

extractDateTimeFrom: aString
        "Answer a String in form 'yy-mm-dd hh:mm:ss'
         describing date and time from DOS directory
         entry aString."
    | answer byte1 byte2 field1 field2 field3 aStream |
    answer := String new: 18.
    aStream := WriteStream on: answer.
        "computing the yy-mm-dd part"
    byte1 := (aString at: 25) asciiValue.
    byte2 := (aString at: 26) asciiValue.
    field1 := (byte2 bitShift: -1) + 80.
    field2 := (byte1 bitShift: -5)
        + ((byte2 bitAnd: 1) * 8).
    field3 := byte1 bitAnd: 31.
    field1 < 10
        ifTrue: [aStream nextPut: $0].
    field1 printOn: aStream.
    aStream nextPut: $-.
    field2 < 10
        ifTrue: [aStream nextPut: $0].
    field2 printOn: aStream.
    aStream nextPut: $-.
    field3 < 10
        ifTrue: [aStream nextPut: $0].
    field3 printOn: aStream.
    aStream space; space.
        "computing the hh:mm:ss part"
    byte1 := (aString at: 23) asciiValue.
    byte2 := (aString at: 24) asciiValue.
    field1 := byte2 bitShift: -3.
    field2 := (byte1 bitShift: -5)
        + ((byte2 bitAnd: 7) * 8).
    field3 := (byte1 bitAnd: 31) * 2.
    field1 < 10
        ifTrue: [aStream nextPut: $0].
    field1 printOn: aStream.
    aStream nextPut: $:.
    field2 < 10
        ifTrue: [aStream nextPut: $0].
    field2 printOn: aStream.
    aStream nextPut: $:.
    field3 < 10
        ifTrue: [aStream nextPut: $0].
    field3 printOn: aStream.
    ^answer!

extractFileNameFrom: aString
        "Answer a string representing the file name
         from a DOS directory entry aString."
    | name ext |
    ((aString at: 12) asciiValue bitAnd: 8) = 8
        ifTrue: [
            ^(aString copyFrom: 1 to: 11)
                trimBlanks asLowerCase].
    name := (aString copyFrom: 1 to: 8) trimBlanks.
    ext := (aString copyFrom: 9 to: 11) trimBlanks.
    ext isEmpty
        ifFalse: [
            name := name, (String with: $.), ext].
    ^name asLowerCase!

extractFlagsFrom: aString
        "Answer a String containing attribute flags
         from a DOS directory entry aString.
         Attributes are: 'r' read only, 'h' hidden,
         's' system, and 'a' archive."
    | answer aByte |
    answer := String new.
    aByte := (aString at: 12) asciiValue.
    (aByte bitAnd: 1) > 0
        ifTrue: [answer := 'r'].
    (aByte bitAnd: 2) > 0
        ifTrue: [answer := answer, 'h'].
    (aByte bitAnd: 4) > 0
        ifTrue: [answer := answer, 's'].
    (aByte bitAnd: 32) > 0
        ifTrue: [answer := answer, 'a'].
    ^answer!

extractSizeFrom: aString
        "Answer the file size extracted from
         a DOS directory entry aString."
    | answer |
    answer := 0.
    32 to: 29 by: -1 do: [ :index |
        answer := answer * 256
            + (aString at: index) asciiValue].
    ^answer!

freeDiskSpace
        "Answer the free space in bytes on the disk
         containing the current directory."
    ^self dosError: (Dos freeDiskSpace)!

pathName: aString
        "Answer a Directory described by
         the complete path name in aString."
    ^self new
        drive: (aString at: 1);
        pathName: (aString copyFrom: 3 to: aString size)!

remove: aString
        "Remove the DOS directory with
         the path name of aString."
    | anInteger |
    anInteger := Dos
        asciizPrimitives: 16r3A00 with: aString asAsciiZ.
    anInteger = -5
        ifTrue: [self dosError: -19]
        ifFalse: [self dosError: anInteger]! !


! OldDirectory methods !

= aDirectory
        "Answer true if aDirectory represents the
         same directory as the receiver, else
         answer false."
    (aDirectory isKindOf: Directory)
        ifFalse: [^false].
    ^self drive == aDirectory drive
        and: [self pathName = aDirectory pathName]!

allSubdirs
        "Answer an OrderedCollection of Arrays for the
         entire subtree of directories under the receiver
         directory. Each Array contains the complete path
         name and the file name of a subdirectory."
    | answer subdirs |
    answer := OrderedCollection new.
    subdirs := self subdirectories.
    answer addAll: subdirs.
    subdirs do: [ :each |
        answer
            addAll:
                (Directory new
                    drive: self drive;
                    pathName: (each at: 1)) allSubdirs].
    ^answer!

contents
        "Answer an OrderedCollection of Strings
         for the receiver Directory.  Each String
         is a DOS directory entry for a file."
    | answer aString fcb anInteger |
    answer := OrderedCollection new.
    fcb := FileControlBlock unopened.
    self makeCurrent.
    [aString := String new: 32.
     (anInteger := self nextEntryInto: aString using: fcb) > 0]
        whileTrue: [
            ((aString at: 12) asciiValue bitAnd: 8) = 8
                ifTrue: [
                    volumeLabel := Directory
                        extractFileNameFrom: aString]
                ifFalse: [
                    (aString at: 1) == $.
                        ifFalse: [answer add: aString]]].
    anInteger = 0
        ifFalse: [^self dosError: anInteger].
    ^answer!

create
    "Create a DOS directory on disk
     for the receiver directory."
    Directory create: (String with: self drive),
        ':', self pathName!

drive
        "Answer the disk drive letter of the receiver."
    ^drive!

drive: aCharacter
        "Initialize the drive for the
         receiver to aCharacter."
    volumeLabel := String new.
    drive := aCharacter!

file: aString
        "Answer a FileStream for the file named
         aString in the current directory.  If the file
         does not exist, it will be created."
    ^FileStream on: (File open: aString in: self)!

file: aString bufferSize: anInteger
        "Answer a FileStream for the file named
         aString in the current directory.  If the file
         does not exist, it will be created."
    ^BufferedFileStream on: (File open: aString in: self)
                        bufferSize: anInteger!

formatted
        "Answer a collection of arrays of file information
         for the receiver directory.  Each array has four
         entries: file name, size, date/time and attributes."
    | answer fileEntries anArray |
    fileEntries := self contents.
    answer := OrderedCollection new: fileEntries size.
    fileEntries do: [ :each |
        (each at: 12) asciiValue = 16
            ifFalse: [
                anArray := Array new: 4.
                anArray
                    at: 1 put:
                        (Directory extractFileNameFrom: each);
                    at: 2 put:
                        (Directory extractSizeFrom: each);
                    at: 3 put:
                        (Directory extractDateTimeFrom: each);
                    at: 4 put:
                        (Directory extractFlagsFrom: each).
                answer add: anArray]].
    ^answer!

makeCurrent
    "Make the receiver directory
     the current DOS directory."
    Directory setTo: (String with: self drive),
        ':', self pathName!

newFile: aString
        "Answer a FileStream for the file named
         aString in the current directory.  If the file
         exists, it will be removed and a new file
         will be created."
    | answer |
    answer := self file: aString.
    answer size > 0
        ifTrue: [
            File remove: answer pathName.
            answer := self file: aString].
    ^answer!

newFile: aString bufferSize: anInteger
        "Answer a FileStream for the file named
         aString in the current directory.  If the file
         exists, it will be removed and a new file
         will be created."
    | answer |
    answer := self file: aString bufferSize: anInteger.
    answer size > 0
        ifTrue: [
            File remove: answer pathName.
            answer := self file: aString bufferSize: anInteger].
    ^answer!

nextEntryInto: aString using: aFileControlBlock
        "Private - Set aString to the next DOS
         directory entry for the receiver.  The
         next entry is described by aFileControlBlock."
    <primitive: 68>
    ^self primitiveFailed!

pathName
        "Answer a String representing the path name
         of the receiver directory (drive letter not
         included)."
    ^pathName trimBlanks!

pathName: aString
        "Set the receiver directory path name to aString."
    aString isEmpty
        ifTrue: [pathName := (String with: $\) asAsciiZ]
        ifFalse: [pathName := aString asAsciiZ]!

pathNameVar
        "Private - Answer the path name
         variable of the receiver."
    ^pathName!

remove
    "Remove the directory described
     by the receiver from the disk."
    Directory remove: (String with: self drive),
        ':', self pathName!

subdirectories
        "Answer an OrderedCollection of arrays, where
         each Array contains the complete path name and
         the file name of a subdirectory of the receiver."
    | answer path name |
    (path := self pathName) = '\'
        ifTrue: [path := String new].
    answer := OrderedCollection new.
    self contents do: [ :entry |
        (entry at: 12) asciiValue = 16
            ifTrue: [
                name := Directory extractFileNameFrom: entry.
                answer add:
                    (Array
                        with: path, '\', name
                        with: name)]].
    ^answer!

subDirectory: pathName
        "Answer a Directory whose path name
         equals the receiver path name
         concatenated with pathName."
    | aString |
    (aString := self pathName) = '\'
        ifTrue: [aString := String new].
    ^self class new
        drive: self drive;
        pathName: aString, '\', pathName!

volumeLabel
        "Answer a String containing the volume label
         for the disk containing the receiver Directory."
    ^volumeLabel! !
