
OrderedCollection variableSubclass: #Process
  instanceVariableNames: 
    'topFrame frameBias '
  classVariableNames: ''
  poolDictionaries: '' !

!Process class methods !

copyStack
        "Answer a Process object containing
         the current stack contents."
    <primitive: 108>!

dropSenderChain
        "Discard stacked message sends (sent but not
         answered) to outermost send, the input request loop."
    <primitive: 109>! !


!Process methods !

contextFor: frameIndex
        "Answer the context object from
         the stack frame at frameIndex."
    ^self frameAt: frameIndex offset: 5!

dropFrame
        "Discard top stack frame."
    | amountToPop frameIndex |
    frameIndex := topFrame + frameBias.
    topFrame := self at: frameIndex + 1.
    amountToPop := (self at: frameIndex + 2)
         + frameBias - 1.
    amountToPop timesRepeat: [self removeFirst].
    frameBias := frameBias - amountToPop!

dropTo: aCompiledMethod
        "Discard all stack entries preceding the
         last occurence of aCompiledMethod."
    | frameIndex lastIndex nextIndex lastStackTop
      topOfStack |
    nextIndex := topFrame + frameBias.
    topOfStack := 1.
    [frameIndex := nextIndex.
    nextIndex := (self at: frameIndex + 1) + frameBias.
    (self at: nextIndex + 1) = 0]
        whileFalse: [
            aCompiledMethod = (self at: nextIndex + 4)
                ifTrue: [
                    lastIndex := frameIndex.
                    lastStackTop := topOfStack].
            topOfStack := (self at: frameIndex + 2)
                + frameBias].
    lastIndex isNil
        ifFalse: [
            topFrame := lastIndex - frameBias.
            startPosition := startPosition
                + lastStackTop - 1.
            frameBias := frameBias - lastStackTop + 1]!

frameAt: frameIndex offset: offset
        "Answer the field of the 'frameIndex + 1'
         stack frame indicated by 'offset'."
    | index |
    index := topFrame.
    frameIndex timesRepeat: [
        index := self at: index + frameBias + 1.
        index = 0
            ifTrue: [
                self error: 'End of stack frame chain']].
    ^self at: index + frameBias + offset!

frameAt: frameIndex offset: offset put: value
        "Answer the field of the 'frameIndex + 1'
         stack frame indicated by 'offset'."
    | index |
    index := topFrame.
    frameIndex timesRepeat: [
        index := self at: index + frameBias + 1.
        index = 0
            ifTrue: [
                self error: 'End of stack frame chain']].
    self at: index + frameBias + offset put: value!

homeFrameOf: frameIndex
        "Answer the stack frame index of the
         associated method frame for the block or
         method frame identified by frameIndex."
    | context homeOffset offset index |
    (context := self contextFor: frameIndex) isNil
        ifTrue: [^frameIndex].
    (homeOffset := context at: 2) = 0
        ifTrue: [^0].  "already exited"
    index := 0.
    offset := topFrame.
    [offset = 0]
        whileFalse: [
            offset := self at: offset + frameBias + 1.
            index := index + 1.
            offset = homeOffset
                ifTrue: [^index]].
    ^0!

methodAt: frameIndex
        "Answer the compiled method for
         the stack frame at frameIndex."
    ^self frameAt: frameIndex offset: 4!

methodAt: frameIndex put: compiledMethod
        "Update the compiled method for
         the stack frame at frameIndex."
    self
        frameAt: frameIndex
        offset: 4
        put: compiledMethod!

receiverAt: frameIndex
        "Answer the receiver from the
         stack frame at frameIndex."
    ^self frameAt: frameIndex offset: 3!

restartAt: frameIndex
        "Restart the process at the method
         described in the stack frame at
         frameIndex."
    | arguments tempCount method receiver index
        instCtr resultPtr |
    method := self methodAt: frameIndex.
    tempCount := method at: 2.
    receiver := self receiverAt: frameIndex.
    resultPtr := self frameAt: frameIndex offset: 2.
    index := resultPtr + frameBias - 1.
    arguments := OrderedCollection new.
    (method at: 3) timesRepeat: [
        index := index - 1.
        arguments addFirst: (self at: index)].
    instCtr := self frameAt: frameIndex offset: 0.
    frameIndex + 1 timesRepeat: [self dropFrame].
    self addFirst: receiver.
    frameBias := frameBias + 1.
    arguments reverseDo: [ :arg |
        self addFirst: arg.
        frameBias := frameBias + 1].
    tempCount < 128
        ifTrue: [
            tempCount + 1 timesRepeat: [
                self addFirst: nil].
            frameBias := frameBias + tempCount + 1]
        ifFalse: [
            self addFirst: (Context new
                method: method
                receiver: receiver
                arguments: arguments
                tempCount: 255 - tempCount
                frame: -5 - frameBias).
            frameBias := frameBias + 1].
    self
        addFirst: method;
        addFirst: receiver;
        addFirst: resultPtr;
        addFirst: topFrame;
        addFirst: instCtr.
    frameBias := frameBias + 5.
    topFrame := 1 - frameBias.
    self
        addFirst: receiver;
        addFirst: 5.
    frameBias := frameBias + 2.
    self swapInAndRestore: false!

resume
        "Resume the receiver process."
    | topFrameIndex currentIC result |
    topFrameIndex := topFrame + frameBias.
    currentIC := self at: topFrameIndex.
    result := self at: topFrameIndex + 3.
    topFrame := self at: topFrameIndex + 1.
    topFrameIndex := topFrameIndex + 5.
    topFrameIndex timesRepeat: [self removeFirst].
    frameBias := frameBias - topFrameIndex.
    self
        addFirst: result;
        addFirst: currentIC.
    frameBias := frameBias + 2.
    self swapInAndRestore: true!

stackOverflow
        "Private - Pop-up a walkback window describing
         the stack overflow condition (too many
         stacked message sends).  This message is
         sent by the virtual machine."
    Debugger new
        resumable: false;
        walkBackFor: self
        label: 'Stack Overflow'!

swapInAndRestore: aBoolean
        "Private - Primitive method to activate the
         receiver process and restore the terminal state."
    <primitive: 104>
    ^self primitiveFailed!

walkBackOn: collector
        "Private - Display the stacked message sends for
         the receiver on the collector stream or TextEditor
         of a window."
    | assoc aClass frameIndex stopSelector |
    stopSelector :=
        collector class == Debugger
            ifTrue: [nil]
            ifFalse: [#Doit].
    frameIndex := 0.
    [(frameIndex := frameIndex + 1) > 10
        or: [(self frameAt: frameIndex offset: 1) = 0]]
        whileFalse: [
            (self homeFrameOf: frameIndex) = frameIndex
                ifFalse: [collector nextPutAll: '[] in '].
            aClass := (self receiverAt: frameIndex) class.
            aClass name first isUpperCase
                ifTrue: [
                    collector nextPutAll: aClass name.
                    assoc := aClass selectorFor:
                        (self methodAt: frameIndex).
                    assoc isNil
                        ifTrue: [
                            collector nextPutAll: '>>-unknown selector-']
                        ifFalse: [
                            assoc value = aClass
                                ifFalse: [
                                    collector
                                        nextPut: $(;
                                        nextPutAll: assoc value name;
                                        nextPut: $)].
                            collector
                                nextPutAll: '>>';
                                nextPutAll: assoc key.
                            assoc key = stopSelector
                                ifTrue: [^collector cr]].
                    collector cr]]! !
