"Smalltalk code from Chapter 10 of the Smalltalk/V Tutorial"!

!Animal methods !

answer: aString
    "Display a message for the receiver animal on
     the reply pane, consisting of the animal's
     class name and name preceding aString."
    habitat answer:
        self class name, ' ', name, ': ', aString!

bounce: distance
        "bounce for distance."
    self answer: 'I am bouncing ', distance printString
        , ' feet'.
    habitat animator
        speed: topSpeed;
        tell: name bounce: distance!

color
        "Answer color."
    ^color!

direction: anAngle
    self answer: 'My direction is now ', anAngle printString
        , ' degrees'.
    habitat animator
        tell: name direction: anAngle!

habitat: aHabitat
         "Change habitat to aHabitat"
     habitat := aHabitat!

home
    self answer: 'I am coming home'.
    habitat animator
        tell: name place: 0@0!

learn: aString action: aBlock
         "Add a pattern of the words in aString to the
         receivers knowledge.  The action to perform
         when the pattern is matched is aBlock."
     | words pattern |
     knowledge isNil
         ifTrue: [knowledge := Dictionary new].
     words := aString asLowerCase asArrayOfSubstrings.
     pattern := Pattern new:
         (Array with: name asLowerCase), words.
     pattern matchBlock: aBlock.
     knowledge at: words put: pattern!

name
        "Answer name."
    ^name!

name: aString
        "Change the receiver animal's name to aString."
    name := aString!

name: aString picture: images color: aColor 
    name := aString.
    picture := images.
    color := aColor!

picture
        "Answer picture."
    ^picture!

reactTo: aWord
         "Send a word to every pattern in knowledge."
     knowledge isNil
         ifTrue: [^self].
     knowledge do: [:pattern | pattern match: aWord]!

reset
         "Reset all patterns in knowledge"
     knowledge isNil
         ifTrue: [^self].
     knowledge do: [:pattern | pattern reset]!

run: distance
        "Run for distance."
    self answer: 'I am running ', distance printString
        , ' feet'.
    habitat animator
        speed: topSpeed;
        tell: name go: distance!

talk
        "Display a message that the receiver can't talk."
    self answer: 'I can''t talk'!

topSpeed: speed
        "Set speed at miles per hour."
    topSpeed := speed.
    self answer: 'My top speed is now ', speed printString!

turn: anAngle
        "Turn direction with anAngle."
    self answer: 'I am turning ', anAngle printString
        , ' degrees'.
    habitat animator
        tell: name turn: anAngle!

walk: distance
        "Walk for distance."
    self answer: 'I am walking ' ,
        distance printString , ' feet'.
    habitat animator
        speed: topSpeed // 2;
        tell: name go: distance! !
                       
!Parrot methods !

talk
        "Display a message containing the receiver
         parrot's vocabulary."
    self answer: vocabulary!

vocabulary: aString
        "Change the receiver parrot's vocabulary
         to aString."
    vocabulary := aString! !
                         
!Dog methods !

bark
        "Have the receiver dog bark by ringing the bell
         and displaying a bark message."
    Terminal bell.
    barksAlot
        ifTrue: [self answer: 'Bow Wow, Bow Wow, Bow Wow!!']
        ifFalse: [self answer: 'Woof']!

beNoisy
        "Change the status of the receiver dog to noisy."
    barksAlot := true.
    self answer: 'I''ll bark a lot'!

beQuiet
        "Change the status of the receiver dog to quiet."
    barksAlot := false.
    self answer: 'I won''t bark much'!

talk
        "Have the receiver dog talk by barking; unless
         barksAlot is nil, in which case the superclass
         can decide how to talk."
    barksAlot isNil
        ifTrue: [super talk]
        ifFalse: [self bark]! !
                       
!AnimalHabitat methods !

add: anAnimal
         "Add anAnimal as an inhabitant of the receiver.
         Notify anAnimal of its new habitat."
     animals isNil
         ifTrue: [animals := Set new].
     animals add: anAnimal.
     anAnimal habitat: self!

animator
        "Answer the animator."
    ^animator!

answer: aString
        "Output aString to the reply Pane."
    (replyStream isNil
        ifTrue: [Transcript]
        ifFalse: [replyStream])
            nextPutAll: aString;
            cr!

graph: aRect
        "Initialize graph pane area to aRect and the
         animation associated with it."
    | aForm |
    aForm := Form
        width: aRect width
        height: aRect height.
    aForm displayAt: aRect origin.  "background"
    animator := Animation new initialize: aRect.
    animals do: [:anAnimal |
        animator
            add: anAnimal picture
            name: anAnimal name
            color: anAnimal color].
    ^aForm!

input
        "Initialize inputPane with inputString."
    ^inputString!

inputMenu
        "Answer a Menu for the input Pane."
    ^Menu
        labels: 'copy\cut\paste\play selection\play all' withCrs
        lines: #(3)
        selectors: #(copySelection cutSelection
            pasteSelection playSelection playAll)!
            
play
         "Play the Script to all of the animals."
     | word |
     Script reset.
     animals do: [ :animal | animal reset ].
     [Script atEnd]
         whileFalse: [
             word := Script next asLowerCase.
             animals do: [ :animal | animal reactTo: word]]!

playAll
        "Accept the entire content of the pane
         as the script and play it to animals."
    self changed: #reply.
    CursorManager execute change.
    self script: inputPane contents.
    self play.
    CursorManager normal change!

playSelection
        "Accept selected string as the script
         and play it to animals."
    self changed: #reply.
    CursorManager execute change.
    self script: inputPane selectedString.
    self play.
    CursorManager normal change!

release
    "Window is closed, release all the
     objects created by this habitat."
    animals isNil
        ifFalse: [
            animals do: [:each |
                each habitat: nil]].
    super release!

reply
        "Initialize reply pane with an
         empty String."
    ^String new!

script: aString
         "Change Script to tbe stream on the
         words in aString."
    | stream word |
    stream := ReadStream on: aString.
    Script := ReadWriteStream on: Array new.
    [(word := stream nextWord) isNil]
        whileFalse:[
            Script nextPut: word]! !
