Transcript cr; show: 'Filing in BitEditor'!

Object subclass: #BitEditor
  instanceVariableNames: 
    'buttons bitPen imagePen imageForm topPane cursorIn scale '
  classVariableNames: ''
  poolDictionaries: 
    'FunctionKeys ' !

!BitEditor class methods ! !


!BitEditor methods !

activatePane
        "The pane of the bit editor has been activated.  Change the cursor
         to the crosshair."
    cursorIn := true.
    CursorManager hair change!

changeBits: aPoint
        "The select button has been pressed at aPoint in the magnified pane.
         Reverse the bit under the cursor and all bits the cursor moves
         over until the select button is released."
    | aChar locX locY newX newY |
    locX := -1.
    locY := -1.
    [newX := Cursor offset x - bitPen clipX // scale.
    newY := Cursor offset y - bitPen clipY // scale.
    (newX = locX and: [newY = locY])
        ifFalse: [
            locX := newX.
            locY := newY.
            self setX: locX
                Y: locY].
    Terminal read == EndSelectFunction]
        whileFalse: [self checkCursor]!

changeStatus: aPoint
        "The select button has been pressed inside the status pane at
         aPoint.  Check to see which button (if any) the cursor was inside
         and take the appropriate action."
    | button oldOffset |
    button := buttons setToPoint: aPoint.
    button = 1
        ifTrue: [
            oldOffset := imageForm offset.
            imageForm fromDisplay: imagePen clipRect.
            ^imageForm offset: oldOffset].
    button = 2
        ifTrue: [
            Terminal initialize.
            ^topPane dispatcher closeIt]!

checkCursor
        "The select button has been released, check if the cursor is still
         in the window and set the cursor shape accordingly."
    | box |
    box := bitPen clipRect.
    (box containsPoint: Cursor offset)
        ifTrue: [
            cursorIn ifFalse: [
                cursorIn := true.
                CursorManager hair change]]
        ifFalse: [
            cursorIn ifTrue: [
                cursorIn := false.
                CursorManager normal change]]!

deactivatePane
        "The pane has been deactivated.  change the cursor to the normal
         cursor shape."
    cursorIn := false.
    CursorManager normal change!

initBits: aRectangle
        "Initialize the bits pane of a BitEditor."
    | aPoint |
    cursorIn := false.
    Display white: aRectangle.
    bitPen clipRect: aRectangle;
        grid: scale.
    bitPen defaultNib: scale - 3 @ (scale - 3);
        combinationRule: Form reverse;
        mask: nil.
    aPoint := 0@0.
    0 to: imageForm height - 1 do: [ :j |
        0 to: imageForm width - 1 do: [ :i |
            aPoint x: i; y: j.
            (self valueAt: aPoint) = 0
                ifTrue: [self setX: i Y: j]]].
    ^Form new extent: aRectangle extent!

initStatus: aRectangle
        "Initialize the status pane (bounded by aRectangle)
         of a BitEditor."
    | size loc i imageSize |
    Display white: aRectangle.
    size := 16@16.
    imageSize := imageForm extent min:
        bitPen clipRect extent + scale // scale.
    i := Font eightLine height * 2 + size y.
    loc := aRectangle width - size x // 2 @
        (aRectangle height - imageSize y
            - (i * 2) // 2).
    buttons :=
        Button new form: Display frame: aRectangle.
    buttons
        loc: (Array with: loc with: 0@i + loc)
        size: size
        label: #('save' 'exit')
        labelLoc: (Array with: loc + (-8 @ size y)
            with: loc + (-8 @ (i + size y)))
        font: Font eightLine.
    loc := aRectangle corner - imageSize.
    Display border: (loc - (3@3) extent: imageSize + (6@6))
        widthRectangle: (2@2 corner: 2@2)
        mask: Form black.
    imagePen := Pen new.
    imagePen clipRect: (loc extent: imageSize).
    imagePen
        combinationRule: Form reverse;
        mask: nil.
    ^Form new extent: aRectangle extent!

openOn: aForm
        "Magnify aForm by a preset scale so that its
         bits can be easily edited and open a bit editor on the
         magnified form."
    | size i bitPane statusBox |
    scale := 8.
    imageForm := aForm.
    size := imageForm extent max: 48@16.
    topPane := TopPane new
        label: 'Bit Editor';
        menu: #windowMenu;
        model: self;
        minimumSize: 1@1;
        yourself.
    topPane addSubpane:
        (GraphPane new
            model: self;
            name: #initStatus:;
            change: #changeStatus:;
            framingBlock: [ :box |
                box origin extent: ((statusBox :=
                    size x + 4) @ box height)]).
    topPane addSubpane:
        (bitPane := GraphPane new
            model: self;
            name: #initBits:;
            change: #changeBits:;
            framingBlock: [ :box |
                box origin + (size x + 4 @ 0)
                extent: box width - statusBox @ box height]).
    i := imageForm width * scale + size x @ (size y * scale) +
        (11 @ (11 + SysFontHeight))
        min: Display extent.
    topPane
        reframe: (Display extent - i // 2 extent: i).
    bitPen := Pen new
        clipRect: bitPane frame.
    topPane dispatcher
        openWindow;
        scheduleWindow!

setX: x Y: y
        "Reverse the bit at x@y in both the magnified form
         and the original form."
    bitPen place:
        x * scale + bitPen clipX + 2
        @ (y * scale + bitPen clipY + 2).
    bitPen copyBits.
    imagePen place:
        x + imagePen clipX @ (y + imagePen clipY).
    imagePen copyBits!

valueAt: aPoint
        "Answer the bit at aPoint in the image (original) form."
    | byteLoc aByte |
    byteLoc :=  aPoint y * imageForm byteWidth + (aPoint x // 8).
    aByte := imageForm bitmap at: byteLoc + 1.
    (aByte bitAnd: (1 bitShift: 7 - (aPoint x \\ 8)))
        = 0
        ifTrue: [^0]
        ifFalse: [^1]!

windowMenu
        "Answer the window menu for the bit editor."
    ^Menu
        labels: 'cycle\close' withCrs
        lines: #()
        selectors: #(cycle closeIt).! !
