"This file contains the Cloner classes. 

 The Cloner does a complete memory walk of all of the objects
 in Smalltalk/V.  It will purge all unused Symbols.  In addition
 it reassigns all of the object pointers to the low end of the 
 object pointer space.

 Note: When you clone, the IMAGE FILE WILL BE REWRITTEN.
 Therefore, you should back up your image file before cloning.
        
 Cloning takes a long time (hours).  To let you know that
 everything is alright, the Cloner writes a dot on the screen
 for every 5 objects cloned.

 To clone Smalltalk/V, make sure that this file is in your
 Disk directory, (the same directory that the image file 
 is in).  Then evaluate the following expression in
 any TextPane, (e.g. the System Transcript):

    (Disk file: 'cloner.st') fileIn; close.

 When cloning is finished Smalltalk/V will exit to DOS.
 You should make a back up of your image file"


Smalltalk 
    removeKey: #Cloner ifAbsent: [];
    removeKey: #CloneSet ifAbsent: [];
    removeKey: #CloneDictionary ifAbsent: [];
    removeKey: #CloneIdentityDictionary ifAbsent: [];
    at: #Clone put: nil !

Set variableSubclass: #CloneSet
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: '' !
 
!CloneSet class methods ! !

!CloneSet methods !

findElementIndex: anObject
    | index indexedObject lastIndex |
    lastIndex := self basicSize.
    index := (Clone hashFor: anObject) \\ lastIndex + 1.
    [(indexedObject := self basicAt: index) = anObject]
        whileFalse: [
            (indexedObject isNil)
                ifTrue: [^index].
            (index := index + 1) > lastIndex
                ifTrue: [ "index wraparound"
                    index := 1]].
    ^index! !

Dictionary variableSubclass: #CloneDictionary
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: '' !

!CloneDictionary class methods ! !

!CloneDictionary methods !

findKeyIndex: key
    | index indexedAssoc lastIndex |
    lastIndex := self basicSize.
    index := (Clone hashFor: key) \\ lastIndex + 1.
    [((indexedAssoc := self basicAt: index) isNil) 
        or: [key = indexedAssoc key]]
            whileFalse: [
                (index := index + 1) > lastIndex
                    ifTrue: [ "index wraparound"
                        index := 1]].
    ^index! !

IdentityDictionary variableSubclass: #CloneIdentityDictionary
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: '' !

!CloneIdentityDictionary class methods ! !

!CloneIdentityDictionary methods !

findKeyIndex: key
    | index lastIndex |
    lastIndex := self basicSize.
    index := ((Clone hashFor: key) bitAnd: lastIndex - 1) + 1.
    [key == (self basicAt: index)]
        whileFalse: [
            (self basicAt: index) isNil
                ifTrue: [^index].
            (index := index + 1) > lastIndex
                ifTrue: [ "index wraparound"
                    index := 1]].
    ^index! !

Object subclass: #Cloner
  instanceVariableNames: 
    'clonePointer clones objectsSeen clonesToBeDumped classesToSkip symbolSetPosition '
  classVariableNames: ''
  poolDictionaries: '' !

!Cloner class methods ! !

!Cloner methods !

addSkipClasses: collection  
        "Add the classes for each symbol
         in collection to cloneClassesToSkip."  
    | class |
    collection do: [ :symbol |                  
        class := Smalltalk at: symbol.
        classesToSkip
            add: class;
            add: class class.
        objectsSeen 
            add: class class]!

cloneFixedPointers
        "Cloning of the system fixed pointers."
    self
        cloneOopFor: false;
        cloneOopFor: 'schedAssocPtr'; 
        cloneOopFor: Character;
        cloneOopFor: nil;
        cloneOopFor: Smalltalk;
        cloneOopFor: Symbol symbolTable;
        cloneOopFor: String;
        cloneOopFor: UndefinedObject;
        cloneOopFor: #value;
        cloneOopFor: #value:;
        cloneOopFor: Symbol;
        cloneOopFor: Context;
        cloneOopFor: Array;
        cloneOopFor: FileControlBlock;
        cloneOopFor: Point;
        cloneOopFor: Message;
        cloneOopFor: true;
        cloneOopFor: CompiledMethod;  
        cloneOopFor: Process;
        cloneOopFor: #mustBeBoolean;
        cloneOopFor: SmallInteger;
        cloneOopFor: #x;
        cloneOopFor: #y;
        cloneOopFor: #stackOverflow;        
        cloneOopFor: Form;
        cloneOopFor: DisplayScreen;
        cloneOopFor: SwappedOutObject;
        cloneOopFor: (Smalltalk
            associationAt: #ErrorCode);
        cloneOopFor: '';  
        cloneOopFor: #doesNotUnderstand:;
        cloneOopFor: #cantReturn;
        cloneOopFor: #controlBreak;
        cloneOopFor: LargeNegativeInteger;
        cloneOopFor: #<;
        cloneOopFor: #>;
        cloneOopFor: #<=;
        cloneOopFor: #=;
        cloneOopFor: #+;
        cloneOopFor: #-;
        cloneOopFor: #*;
        cloneOopFor: #//;
        cloneOopFor: #\\;
        cloneOopFor: #bitAnd:;
        cloneOopFor: #bitShift:;
        cloneOopFor: #@;                                    
        cloneOopFor: LargePositiveInteger;                    
        cloneOopFor: #userPrimMissing;
        cloneOopFor: #startUp.
    0 to: 255 do: [ :i | self cloneOopFor: i asCharacter]!

cloneOopFor: anObject
        "Answer the cloned oop for anObject."
    anObject isNil
        ifTrue: [
            clonePointer = 7
                ifTrue: [ 
                    clonesToBeDumped addLast: anObject.
                    clonePointer := 9].
            ^7 ].
    anObject class == SmallInteger
        ifTrue: [^anObject * 2].
    ^clones at: anObject
        ifAbsent: [
            (classesToSkip includes: anObject class)
                ifTrue:  [^clones at: anObject put: 7]
                ifFalse: [ 
                    clones at: anObject put: clonePointer.
                    clonePointer := clonePointer + 2.
                    clonesToBeDumped addLast: anObject.
                    ^clonePointer - 2]]!

cloneSystemOn: aStream
        "Clone the system on aStream." 
    | newSymbolSet count |  
    count := 0.
    Clone := self.
    self initializeClones.                         
    self skipClasses.
    aStream                     "set the first four bytes to"
        nextTwoBytesPut: 4;     "indicate the position at which"
        nextTwoBytesPut: 0.     "the image data starts"
    aStream 
        nextTwoBytesPut: 16r20. "Smalltalk/V R1.0"
    aStream 
        nextTwoBytesPut: 0.     "reserve the first two bytes into"
                                "which ClonePointer will be placed"
                                "after the entire dump"
    objectsSeen do: [ :object | 
        clones at: object put: 7 ].
    self cloneFixedPointers. 
        "reserving 3000 free oops"
    clonePointer := clonePointer + 6000.
    clonesToBeDumped add: 6000.  
    self cloneSelectors.
    [clonesToBeDumped isEmpty]
        whileFalse: [  
            (count := count + 1 \\ 5) == 0
                ifTrue: [Terminal nextPut: $.]. 
            clonesToBeDumped removeFirst 
                dumpOn: aStream for: self].
    clonesToBeDumped := nil.
    aStream position: symbolSetPosition.
    newSymbolSet := SymbolSet new: Symbol symbolTable basicSize.    
    clones keysDo: [ :each |
        each class == Symbol
            ifTrue: [newSymbolSet add: each]].
    self dump: newSymbolSet on: aStream.
    aStream 
        position: 6;
        nextTwoBytesPut: clonePointer;
        flush.
    Symbol symbolTable do: [ :each |
        (newSymbolSet includes: each)
            ifFalse: [
                Transcript nextPutAll: each; cr]].
    clonePointer := clones := objectsSeen := nil.
    clonesToBeDumped := classesToSkip := nil!

dump: anObject on: aStream
        "anObject dumped as a stream of bytes"
    | index |
    aStream nextTwoBytesPut: 
        (self cloneOopFor: anObject class).
    index := anObject class instSize + anObject basicSize.
    aStream nextTwoBytesPut: index + 1 * 2.
    1 to: index do: [ :i |
        aStream nextTwoBytesPut: 
            (self cloneOopFor: (anObject instVarAt: i))]!

dump: anObject on: aStream for: aClass
        "Dumped anObject as a stream of bytes
         changing the class to aClass"
    | index |
    aStream nextTwoBytesPut: (self cloneOopFor: aClass).
    index := anObject class instSize + anObject basicSize.
    aStream nextTwoBytesPut: index + 1 * 2.
    1 to: index do: [ :i |
        aStream nextTwoBytesPut: 
            (self cloneOopFor: (anObject instVarAt: i))]!

hashFor: anObject    
        "Hash using cloned object pointer if appropriate"
    | hash |
    (hash := anObject hash) = anObject basicHash
        ifTrue:  [^(self cloneOopFor: anObject) // 2 bitAnd: 16r3FFF]
        ifFalse: [
            ((anObject isKindOf: IndexedCollection)
                and: [anObject class ~~ String])
                    ifTrue: [
                        ^(anObject size = 0 
                          or: [anObject == (anObject at: 1)])
                            ifTrue: [7]
                            ifFalse: [self hashFor: (anObject at: 1)]]
                    ifFalse: [^hash]]!

initializeClones
        "Initialize the cloning database"
    clonePointer := 1.
    clones := IdentityDictionary new.   
    objectsSeen := Set new.
    clonesToBeDumped := OrderedCollection new. 
    classesToSkip := Set new.
    self addSkipClasses: (Array 
        with: #Cloner 
        with: #CloneSet 
        with: #CloneDictionary
        with: #CloneIdentityDictionary)!

newCloneDictionary: anInteger
    ^CloneDictionary new: anInteger!

newCloneIdentityDictionary: anInteger
    ^CloneIdentityDictionary newSize: anInteger!

newCloneSet
    ^CloneSet new!

skipClasses
    classesToSkip do: [ :each |
        each superclass subclasses: 
            (each superclass subclasses asOrderedCollection 
                remove: each; 
                yourself)]!

symbolSetPosition: anInteger
    symbolSetPosition := anInteger!
                                          
cloneSelectors           
    | classes selectors |
    selectors := Set new: 1000.
    classes := Object allSubclasses.
    classes add: Object.
    classes addAll: Object class allSubclasses.
    classes add: Object class.
    classes do: [:class |
        (classesToSkip includes: class)
            ifFalse: [
                selectors addAll: 
                    class methodDictionary keys]].
    selectors := selectors asOrderedCollection.
    [selectors size > 0]
        whileTrue: [
            self cloneOopFor:
                selectors removeFirst.
            selectors size > 0
                ifTrue: [
                    self cloneOopFor:
                        selectors removeLast]]! !

"the following expression does the actual cloning"
| notSwappedHeader image imageHeader file |
Smalltalk isFloppy
    ifTrue: [
        Prompter
            prompt: 'insert image diskette into device ',
                (String with: Disk drive asUpperCase),
                ':  and press return'
            default: ''].
notSwappedHeader := String
    with: 4 asCharacter
    with: 0 asCharacter   
    with: 0 asCharacter
    with: 0 asCharacter.
image := Disk file: 'image'.
imageHeader := image copyFrom: 1 to: 4.
image close.
imageHeader = notSwappedHeader
    ifTrue: [        
        file := Disk newFile: 'image'.
        Cloner new cloneSystemOn: file.
        file close]
    ifFalse: [
        file := Disk newFile: 'newimage'.
        Cloner new cloneSystemOn: file.
        file close.
        File remove: (Disk file: 'image') pathName.
        File
            rename: file pathName
            to: (String with: Disk drive with: $:),
                (Disk pathName = '\'
                    ifTrue: ['']
                    ifFalse: [Disk pathName]), '\image'].
    Smalltalk exit: true !

