Transcript cr; show: 'Filing in Dashboard'!

ControlPanel subclass: #Dashboard
  instanceVariableNames: 
    'idleSpeed rpmToOil gearRatios accToRpm rpmToFuel timeToTemp clockOn clockString '
  classVariableNames: ''
  poolDictionaries: 
    'FunctionKeys CharacterConstants ' !

!Dashboard class methods !

new
    ^super new open! !


!Dashboard methods !

accelerator
        "Change accelerator."
    (dials at: #accelerator) changeReading:
        ((dials at: #rpm) reading - idleSpeed
        * (1 / accToRpm)) rounded!

changed: aSymbol
        "The dial denoted by aSymbol has been changed,
         notify the other dials then check the temp
         and fuel dials."
    super changed: aSymbol.
    (dials at: aSymbol) lastChange = 0
        ifFalse: [
            self
                checkFuel;
                checkOverheat]!

checkFuel
        "Check whether the fuel is out."
    | aString aDial |
    aDial := dials at: #fuel.
    aDial reading > 0
        ifTrue: [^self].
    aString := 'I''m out of fuel'.
    (GraphPane notifier: 'W A R N I N G !!'
        content: aString
        at: aDial frame center x - (aString size * SysFontWidth
            * 3 // 2) @ aDial frame bottom
        menu: (Menu labels: 'continue\fillup' withCrs
            lines: #()
            selectors: #(continue fillup))) == #fillup
        ifTrue: [
            (dials at: #fuel) changeReading: 20]!

checkOverheat
        "Check whether it's overheated when
         the temperature dial is changed."
    | aString aDial |
    aDial := dials at: #temp.
    aDial reading > 211
        ifFalse: [^self].
    aString := 'I''m overheating'.
    (GraphPane notifier: 'W A R N I N G !!'
        content: aString
        at: aDial frame center x - (aString size * SysFontWidth
            * 3 // 2) @ aDial frame bottom
        menu: (Menu labels: 'continue\cool down' withCrs
            lines: #()
            selectors: #(continue cool))) == #cool
        ifTrue: [
            (dials at: #temp) changeReading: 150]!

clock
        "Toggle the dashboard clock on or off."
        | aRect |
    clockOn
        ifTrue: [
            Time clockTicksOff.
            aRect := (dials at: #temp) frame.
            Display white:
                (aRect center x - 32 @ (aRect bottom + 16)
                    extent: 64 @ SysFontHeight).
             clockOn := false]
        ifFalse: [
            Time clockTickPeriod: 18.
            clockOn := true]!

clockEvent: anInteger
        "Clock interrupt sensed, update
         time related apparatus for
         anInteger clock interrupts."
    | dial aStream |
    dials do: [ :aDial |
        aDial lastChange: 0].
    (dials at: #fuel) increment:
        0 - ((dials at: #rpm) reading * rpmToFuel
        * anInteger / 60).
    (dial := dials at: #temp) reading <= 150 ifTrue: [
        dial increment: timeToTemp].
    aStream := WriteStream on: (String new: 8).
    Time now printOn: aStream.
    aStream contents displayAt:
        dial frame center x - 32 @ (dial frame bottom + 16).
    self checkFuel.
    self checkOverheat!

close
        "Close the model. Turn
         off the clock interrupt."
    Time clockTicksOff!

fuel
        "Change fuel reading."!

gearReading
        "Answer current gear."
    ^#(0 1 4 3 2 1) at: (dials at: #gear) selection!

initDial: aSymbol
    in: aRectangle
    labeled: aString
    marks: markArray
    scaled: anInterval
    dependents: anArray
        "Initialize dial named aSymbol within
         aRectangle on Display as specified."
    | center font |
    font := Font eightLine.
    center := aRectangle center +
        (0 @ (font height * Aspect) truncated).
    dials at: aSymbol put:
        (Dial new
            form: Display
                frame: aRectangle;
            drawDial: aRectangle
                center: center
                radius: (aRectangle width // 2
                    min: aRectangle height // Aspect // 2)
                    - (font height // Aspect)
                lowReading: anInterval first
                highReading: anInterval last
                lowDegree: 140
                highDegree: 40
                scales: anInterval increment
                font: font
                marks: markArray;
            paintLabel: aString
                labelLoc: center x - (aString size / 2 * font width)
                     rounded @ (center y + 2 +
                    (font height * 3 * Aspect) truncated)
                labelFont: font;
            dependentArray: anArray;
            yourself)!

initGauge: aSymbol
    in: aRectangle
    labeled: aString
    marks: markArray
    scaled: anInterval
    dependents: anArray
        "Initialize the dial aSymbol within
         aRectangle on Display as specified."
    | center height font |
    font := Font eightLine.
    height := aRectangle height - (font height) - 4.
    center := (aRectangle width // 2 @ height)
        + aRectangle origin.
    dials at: aSymbol put:
        (Dial new
            form: Display
                frame: aRectangle;
            drawDial: (aRectangle origin
                    corner: aRectangle corner x @ center y)
                center: center
                radius: (aRectangle width // 2
                    min: height // Aspect) - 20
                lowReading: anInterval first
                highReading: anInterval last
                lowDegree: 200
                highDegree: 340
                scales: anInterval increment
                font: font
                marks: markArray;
            paintLabel: aString
                labelLoc: center x - (2 * font width) @
                    (center y + 4)
                labelFont: font;
            yourself)!

initGear: aRectangle on: aForm
        "Initialize gear buttons within
         aRectangle on aForm."
    | center font aButton locArray labLocArray labPoint exPoint|
    font := Font eightLine.
    center := aRectangle width - (font width * 2 * 6) // 2
        @ (aRectangle height // 2 - font height - 4).
    aButton := Button new form: aForm frame: aRectangle.
    exPoint := font width * 2 @ (font height * 2).
    locArray := Array new: 6.
    labLocArray := Array new: 6.
    labPoint := center + (font width // 2 @ (font height // 2)).
    1 to: 6 do: [ :i |
        locArray at: i put:
            center + ((i - 1 * exPoint x) @ 0).
        labLocArray at: i put:
            labPoint + ((i - 1 * exPoint x) @ 0)].
    dials at: #gear put:
        (aButton
            loc: locArray
                size: exPoint
                label: #('P' 'R' 'N' '3' '2' '1')
                labelLoc: labLocArray
                font: font;
            paintLabel: 'Gear'
                labelLoc: center + aRectangle origin + (exPoint x * 2
                    @ (exPoint y + (font height // 2)))
                labelFont: font;
                dependentArray: #(mph);
            selection: 1)!

initPedal: aSymbol
    in: aRectangle
    labeled: aString
    dependents: anArray
        "Initialize a dial named aSymbol as specified."
    | center r font rect |
    font := Font eightLine.
    rect := aRectangle origin extent:
        aRectangle width @
        (aRectangle height - (font height * 2)).
    r := (rect width min:
        rect height // Aspect) - 4.
    center := rect left + (rect width - r // 2)
        @ rect bottom.
    dials at: aSymbol put:
        (Dial new
            form: Display
                frame: aRectangle;
            framelessDial: aRectangle
                center: center
                radius: r
                lowReading: 0
                highReading: 30
                lowDegree: 315
                highDegree: 360
                scales: 30/9
                font: font
                marks: #('0' '1' '2' '3');
            paintLabel: aString
                labelLoc: center x - 4 + ((r - (11 * font width) max: 0)
                     // 2) @
                    (center y + (font height) - 2)
                labelFont: font;
            dependentArray: anArray).!

menu
        "Private - Answer the graphics pane menu."
        | aString |
    clockOn
        ifTrue: [aString := 'clock on -> off']
        ifFalse: [aString := 'clock off -> on'].
    ^Menu
        labels: aString
        lines: Array new
        selectors: #(clock)!

mph
        "Change mph reading."
    | dial gear |
    gear := self gearReading.
    dial := dials at: #mph.
    gear == 4            "neutral"
        ifTrue: [^self].
    gear = 0             "park"
        ifTrue: [^dial changeReading: 0].
    dial changeReading:
        ((dials at: #rpm) reading *
        (gearRatios at: gear)) rounded!

oil
        "Change oil reading."
    (dials at: #oil) changeReading:
        ((dials at: #rpm) reading *
            rpmToOil) rounded!

open
        "Open a dashboard for Chevrolet
         and initialize coefficients."
    idleSpeed := 10.
    accToRpm := 80 / 30.
    rpmToOil := 90 / 90.
    gearRatios := Array
        with: 40/90 with: 60/90 with: 100/90.
    rpmToFuel := 1 / 20.
    timeToTemp := 40 / 60.
    self open: 'Chevrolet'!

panel: box
        "Initialize and draw all of the dials and gauges of the dashboard.
         Answer the back up form for the dashboard."
    Display fill: box
        rule: Form over
        mask: Form white.
    dials := Dictionary new: 7.
    self
        initGauge: #fuel
            in: (box scaleTo: (0@0 extent: 1/3 @ (1/4)))
            labeled: 'Fuel'
            marks: #('E' '1/2' 'F')
            scaled: (0 to: 20 by: 20/16)
            dependents: #();
        initGauge: #temp
            in: (box scaleTo: (1/3 @ 0 extent: 1/3 @ (1/4)))
            labeled: 'Temp'
            marks: #('C' 'N' 'H')
            scaled: (0 to: 300 by: 300/8)
            dependents: #();
        initGauge: #oil
            in: (box scaleTo: (2/3 @ 0 extent: 1/3 @ (1/4)))
            labeled: 'Oil'
            marks: #('0' '30' '60' '90')
            scaled: (0 to: 90 by: 10)
            dependents: #();
        initDial: #mph
            in: (box scaleTo: (0 @ (1/4) extent: 1/2 @ (1/2)))
            labeled: 'mph'
            marks: #('0' '20' '40' '60' '80' '100')
            scaled: (0 to: 140 by: 140/50)
            dependents: #(rpm);
        initDial: #rpm
            in: (box scaleTo: (1/2 @ (1/4) extent: 1/2 @ (1/2)))
            labeled: 'rpm/1K'
            marks: #('0' '2' '4' '6' '8' '10')
            scaled: (0 to: 100 by: 4)
            dependents: #(mph oil accelerator);
        initPedal: #accelerator
            in: (box scaleTo: (0 @ (3/4) extent: 1/3 @ (1/4)))
            labeled: 'Accelerator'
            dependents: #(rpm);
        initPedal: #brake
            in: (box scaleTo: (1/3 @ (3/4) extent: 1/3 @ (1/4)))
            labeled: 'Brake'
            dependents: #(rpm);
        initGear: (box scaleTo: (2/3 @ (3/4) extent: 1/3 @ (1/4)))
            on: Display.
    (dials at: #fuel) changeReading: 20.
    clockOn := false.
    ^Form new
        width: box width
        height: box height!

rpm
        "Change rpm reading."
    | dial gear |
    gear := self gearReading.
    dial := dials at: #rpm.
    (gear == 0 or: [gear == 4 or: [(dials at: #mph) lastChange = 0]])
        ifTrue: [     "park/neutral"
            dial changeReading:
                ((dials at: #accelerator) reading
                - (dials at: #brake) reading
                * accToRpm) rounded + idleSpeed]
        ifFalse: [
            dial changeReading:
                ((dials at: #mph) reading *
                (1 / (gearRatios at: gear))) rounded]!

topMenu
        "Answer the window menu."
    ^Menu
        labels: 'cycle\close' withCrs
        lines: #()
        selectors: #(cycle closeIt).! !
