Transcript cr; show: 'Filing in Dial'!

Scale subclass: #Dial
  instanceVariableNames: 
    'r center tip lowDegree highDegree readingsPerDegree pen '
  classVariableNames: ''
  poolDictionaries: '' !

!Dial class methods ! !


!Dial methods !

arcTan90: aPoint
        "Compute the angle of the vector from
         0@0 to aPoint in the interval 0 to 90
         degrees (adjusting for the screen
         aspect ratio).  This method does not use
         floating point."
    | ratio tan x y |
    (y := aPoint y) = 0 ifTrue: [^0].
    (x := aPoint x) = 0 ifTrue: [^90].
    y := y / Aspect.
    ratio := y abs * 100 // x abs.
    1 to: 89 do: [ :i |
        tan := i integerSin * 100 // i integerCos.
        ratio <= tan
            ifTrue: [^i]].
    ^90!

arcTan: aPoint
        "Compute the angle of the vector from
         0@0 to aPoint in the interval 0 to 360
         degrees (adjusting for the screen
         aspect ratio).  This method does not use
         floating point."
    | degree |
    degree := self arcTan90: aPoint.
    aPoint x < 0
        ifTrue: [
            aPoint y < 0
                ifTrue: [^180 + degree]
                ifFalse: [^180 - degree]]
        ifFalse: [
            aPoint y < 0
                ifTrue: [^360 - degree]
                ifFalse: [^degree]]!

degreeToReading: degreeInteger
        "Convert degreeInteger to the receiver
         dial reading."
    | newDegree maxRel |
    maxRel := self relativeDegree: highDegree.
    newDegree := self relativeDegree: degreeInteger.
    newDegree > maxRel
        ifTrue: [
            newDegree > (maxRel // 2 + 180)
                ifTrue: [newDegree := 0]
                ifFalse: [newDegree := maxRel]].
    ^(newDegree * readingsPerDegree) rounded!

drawDial: clipRectangle
        "Draw the curved outline of the receiver dial on the form."
    pen mask: foreColor.
    clipRectangle bottom = center y
        ifTrue: [
            pen place: center;
                up;
                direction: 180;
                go: r;
                down;
                direction: 0;
                go: r * 2].
    pen place: center;
        clipRect: clipRectangle;
        ellipse: r aspect: 1!

drawDial: clipRectangle
    center: aPoint
    radius: rInteger
    lowReading: lowReadInteger
    highReading: highReadInteger
    lowDegree: lInteger
    highDegree: hInteger
    scales: scaleInteger
    font: aFont
    marks: anArray
        "Draw the dial, its tick marks, and labels."
    | scanner |
    center := aPoint.
    r := rInteger.
    lowDegree := lInteger.
    highDegree := hInteger.
    lowReading :=
    reading := lowReadInteger.
    highReading := highReadInteger.
    readingsPerDegree := highReading - lowReading /
        (self relativeDegree: highDegree).
    scanner := CharacterScanner new.
    scanner
        initialize: clipRectangle
            font: aFont
            dest: form;
        setForeColor: foreColor
            backColor: backColor.
    pen := (Pen new initPen: form)
        defaultNib: 2@2.
    self drawDial: clipRectangle.
    pen defaultNib: 1@1.
    self
        drawTickMarks: scanner
            scale: scaleInteger
            letters: anArray;
        drawNeedle!

drawNeedle
        "Draw the needle of the receiver at the current reading."
    pen down;
        defaultNib: 2@2;
        place: center;
        mask: foreColor;
        direction: (reading - lowReading
            // readingsPerDegree) rounded + lowDegree;
        go: (tip := (r - 18 max: 2))!

drawTickMarks: scanner scale: anInteger letters: anArray
        "Draw the recievers tick marks and labels."
    | degree tickSize markInc unitsPerScale
      aString x y marks |
    marks := (highReading - lowReading / anInteger) rounded.
    tickSize := 9.
    unitsPerScale := (self relativeDegree: highDegree)
        / marks.
    markInc := marks // (anArray size - 1).
    degree := lowDegree.
    1 to: marks + 1 do: [:i |
        pen direction: degree rounded;
            place: center;
            up;
            go: r - tickSize - 4.
        i \\ markInc = 1
        ifTrue: [
            pen down; go: tickSize + 2;
                up; go: 4.
            degree >= 180
                ifTrue: [
                    y := pen location y - scanner font height + 2]
                ifFalse: [
                    y := pen location y - 2].
            x := pen location x.
            aString :=
                anArray at: i // markInc + 1.
            (degree == 270 or: [degree == 90])
                ifTrue: [
                    x := x - (aString size
                        * scanner font width // 2).
                    y := y - 2]
                ifFalse: [
                    (degree < 270 and: [degree > 90])
                        ifTrue: [x := x - (aString size
                            * scanner font width) - 3]
                        ifFalse: [x := x + 3]].
            scanner display: aString
                from: 1
                to: aString size
                at: x @ y - frame origin]
        ifFalse: [
            pen go: 4;
                down;
                go: tickSize - 4].
        degree := degree + unitsPerScale.
        degree := degree - ((degree // 360) * 360)]!

eraseNeedle
        "Erase the current dial reading."
    pen down;
        defaultNib: 2@2;
        mask: backColor;
        place: center;
        go: tip!

framelessDial: clipRectangle
    center: aPoint
    radius: rInteger
    lowReading: lowReadInteger
    highReading: highReadInteger
    lowDegree: lInteger
    highDegree: hInteger
    scales: scaleInteger
    font: aFont
    marks: anArray
        "Draw the receiver dial, its tick marks, and labels without drawing
         the curved border of the dial."
    | scanner |
    center := aPoint.
    r := rInteger.
    lowDegree := lInteger.
    highDegree := hInteger.
    lowReading :=
    reading := lowReadInteger.
    highReading := highReadInteger.
    readingsPerDegree := highReading - lowReading /
        (self relativeDegree: highDegree).
    scanner := CharacterScanner new.
    scanner
        initialize: clipRectangle
            font: aFont
            dest: form;
        setForeColor: foreColor
            backColor: backColor.
    pen := (Pen new: form)
        clipRect: clipRectangle;
        defaultNib: 1@1.
    self
        drawTickMarks: scanner
            scale: scaleInteger
            letters: anArray;
        drawNeedle!

relativeDegree: degreeInteger
        "Answer the total degrees between
         degreeInteger and lowDegree."
    | degree |
    degreeInteger < lowDegree
        ifTrue: [
            ^360 - lowDegree + degreeInteger]
        ifFalse: [^degreeInteger - lowDegree]!

setToDegree: anInteger
        "Set the reading of the receiver
         dial to anInteger degrees."
    self changeReading: lowReading +
        (self degreeToReading: anInteger)!

setToPoint: aPoint
        "Set the reading of the receiver
         dial so the line defined by the
         needle goes through aPoint."
    self setToDegree:
        (self arcTan: aPoint - form offset - center)! !
