"This file installs the File Browser application.  To install
 this application evaluate the following expression (note that since
 there is no '\' character in the path name, the file must be in the
 current directory):

    (File pathName: 'filbrwsr.st') fileIn; close.

 The file filbrwsr.doc contains a brief explanation and user's
 guide.

 To open a File Browser, evaluate the following expression:

    FileBrowser new openOn: (File pathName: 'fileName.ext')

"!

Object subclass: #FileBrowser
  instanceVariableNames: 
    'fileStream filePane '
  classVariableNames: 
    'FileMenu '
  poolDictionaries: '' !

!FileBrowser methods !

accept
        "write the file pane contents out as the
         file stream contents"
    filePane dispatcher modified
        ifTrue: [
            filePane fileOutOn: fileStream.
            filePane dispatcher modified: false]!

cancel
        "write the file pane contents out as the
         file stream contents"
    filePane fileInFrom: fileStream.
    filePane dispatcher modified: false!

file
        "Private - Read the selected file and display
         its contents in the text pane."
    CursorManager execute change.
    filePane fileInFrom: fileStream.
    fileStream close.
    filePane dispatcher modified: false.
    filePane forceSelectionOntoDisplay.
    CursorManager normal change!

fileMenu
        "answer the menu for editing"
    ^Menu
        labels: 'restore\install\print\copy\cut\paste\show it\do it\save\next menu' withCrs
        lines: #(3 6 9)
        selectors: #(file fileItIn print copySelection cutSelection
                     pasteSelection printIt doIt accept misc)!

openOn: aFileStream
        "open a window on the file stream for editing"
    | topPane |
    fileStream := aFileStream.
    filePane := TextPane new
        model: self;
        name: #file;
        menu: #fileMenu.
    topPane := TopPane new.
    topPane
        label: aFileStream pathName, ' | File Editor';
        addSubpane: filePane.
    topPane dispatcher open scheduleWindow!

release
    fileStream notNil
        ifTrue: [
            (fileStream == (Sources at: 2) or: [
                    fileStream == (Sources at: 1)])
                ifFalse: [fileStream close].
            fileStream := nil].
    super release! !

!DiskBrowser methods !

fileListMenu
        "Private - Answer the file pane menu."
    ^Menu
        labels: 'remove\rename\copy\print\create\mode\browse'
             withCrs
        lines: Array new
        selectors: #(removeFile renameFile copyFile printFile
            createFile changeFileMode browseFile) !

browseFile
        "Open a File Browser on selected file"
    | aString |
    selectedFile isNil
        ifTrue: [^nil].
    (aString := selectedDirectory pathName) = '\'
        ifTrue: [aString := String new].
    (Smalltalk at: #FileBrowser
        ifAbsent: [self error: 'FileBrowser class is missing'])
            new openOn:
                (File pathName:  (String with: device), ':',
                 aString, '\',  selectedFile) ! !

