Transcript cr; show: 'Filing in FreeDrawing'!

Pen subclass: #FreeDrawing
  instanceVariableNames: 
    'state nibSize nibMask nibRule scanner pane bitForm menu lastCursor '
  classVariableNames: 
    'PictureDictionary '
  poolDictionaries: 
    'FunctionKeys ' !

!FreeDrawing class methods !

new
        "Prompt the user for a rectangle and open an instance of the
         receiver in it."
    PictureDictionary isNil
        ifTrue: [PictureDictionary := Dictionary new].
    ^super new open!

pictureDictionary
        "Answer the PictureDictionary."
    ^PictureDictionary! !


!FreeDrawing methods !

activatePane
        "The drawing pane has been activated, restore the prior
         cursor shape."
    lastCursor change!

backspace
        "Erase the charater preceding the cursor."
    Cursor offset:
        Cursor offset -
        ((scanner font charWidth: $ ) @ 0).
    scanner
        display: ' '
        from: 1
        to: 1
        at: Cursor offset - scanner frame origin!

bitEdit
        "Prompt the user for region of the form to magnify, and open
         a bit editor on it."
    | loc ext |
    pane deactivatePane.
    loc := PointDispatcher pointFromUserDisplaying:
        (Cursor offset extent:
        (ext := Display width - 6 // 9 @
            (Display height - 26 // 8)) - 1).
    BitEditor new openOn:
        (bitForm := Form fromDisplay: (loc extent: ext))!

changeFont
        "Prompt the user for a font and make it the drawing font."
    | aFont |
    aFont := (Menu
        labels: 'eight line\fourteen line\sixteen line' withCrs
        lines: Array new
        selectors: #( eightLine fourteenLine sixteenLine))
            popUpAt: Cursor offset.
    aFont isNil
        ifFalse: [scanner setFont: (Font perform: aFont)]!

changeMask
        "Prompt the user for a new pen color for drawing (mask)."
    | choice |
    choice := (Menu
        labels: 'white\light gray\gray\dark gray\black\other' withCrs
        lines: #(5)
        selectors: #(white lightGray gray darkGray black other))
            popUpAt: Cursor offset.
    choice isNil
        ifTrue: [^self].
    choice = #other
        ifTrue: [
            choice := Prompter prompt: 'Pen mask expression (a Form)'
            defaultExpression: 'Form black'.
            choice isNil
                ifTrue: [^self].
            nibMask := choice]
        ifFalse: [nibMask := Form perform: choice].
    nibMask class == Form
        ifTrue: [self mask: nibMask]!

changeRule
        "Prompt the user for a new drawing rule (used by BitBlt)."
    | aRule |
    aRule := (Menu
        labels: 'over\opaque\under' withCrs
        lines: Array new
        selectors: #(over orThru under)) popUpAt: Cursor offset.
    aRule isNil
        ifTrue: [^self].
    self combinationRule: (nibRule := Form perform: aRule)!

changeSize
        "Prompt the user for a new drawing nib size."
    | nib |
    nib := Prompter prompt: 'Pen nib size (x @ y)'
        defaultExpression: ' 1 @ 1 '.
    nib isNil
        ifFalse: [
            nibSize := nib.
            self defaultNib: nibSize]!

circle
        "Change the drawing state to circle drawing."
    self newState: #circle:.
    CursorManager leftArrow change!

circle: aCharacter
        "Track the cursor with a circle until the user releases the
         select button.  The circle's center is defined by the point
         at which the select button is pressed and the radius by the
         point at which it is released.  Draw the circle upon completion."
    | aChar radius center |
    aChar := aCharacter.
    [aChar == SelectFunction]
        whileFalse: [aChar := Terminal read].
    self defaultNib: 1@1;
        mask: Form white;
        combinationRule: Form reverse;
        place: (center := Cursor offset);
        ellipse: (radius := 0) aspect: 1.
    CursorManager rightArrow change.
    [Terminal read == EndSelectFunction]
        whileFalse: [
            self ellipse: radius aspect: 1;
                place: center;
                ellipse: (radius := (center x - Cursor offset x) abs)
                    aspect: 1.
            Cursor offset: Cursor sense x @ center y].
    CursorManager leftArrow change.
    self ellipse: radius aspect: 1;
        reset;
        ellipse: radius aspect: 1!

clear
        "Erase the drawing form to white."
    pane form fill: Form white.
    pane showWindow!

copyGraph
        "Prompt the user for a name and a rectangle.  Copy the area
         inside the rectangle into a form stored under the name in
         the GraphDictionary."
    | aName names newForm |
    names := PictureDictionary keys asSortedCollection asArray,
        (Array with: 'new name').
    aName := (Menu
        labelArray: names
        lines: Array new
        selectors: names)
            popUpAt: Cursor offset.
    aName = 'new name'
        ifTrue: [
            aName := Prompter prompt: 'Graph object name'
            default: 'aGraph'].
    aName isNil ifTrue: [^self].
    newForm := Form fromUser.
    PictureDictionary at: aName put: newForm!

deactivatePane
        "The pane is deactivating, remember the cursor shape."
    lastCursor := Cursor.
    CursorManager normal change!

draw
        "Initiate free hand drawing."
    self newState: #draw:.
    CursorManager normal change!

draw: aCharacter
        "Follow the cursor while drawing until the select button
         is released."
    | aChar |
    aChar := aCharacter.
    [aChar == SelectFunction]
        whileFalse: [aChar := Terminal read].
    self place: Cursor offset.
    [Terminal read == EndSelectFunction]
        whileFalse: [self goto: Cursor offset]!

ellipse
        "Initiate ellipse drawing."
    self newState: #ellipse:.
    CursorManager leftArrow change!

ellipse: aCharacter
        "Follow the cursor with an ellipse until the select button is
         released, then draw the ellipse.  The ellipse aspect ration is
         defined by the point at which the select button is pressed and the
         point at which it is released."
    | aChar radius aspect center |
    aChar := aCharacter.
    [aChar == SelectFunction]
        whileFalse: [aChar := Terminal read].
    self defaultNib: 1@1;
        mask: Form white;
        combinationRule: Form reverse;
        place: (center := Cursor offset);
        ellipse: (radius := 0) aspect: (aspect := 1).
    CursorManager upArrow change.
    [Terminal read == EndSelectFunction]
        whileFalse: [
            self ellipse: radius aspect: aspect;
                place: center;
                ellipse: (radius := (Cursor offset x - center x) abs)
                    aspect: (aspect := Cursor offset y
                        - center y / ((radius * Aspect) max: 1))].
    CursorManager leftArrow change.
    self ellipse: radius aspect: aspect.
    self reset;
        ellipse: radius aspect: aspect!

erase
        "Change the drawing pen color to white."
    self newState: #draw:.
    self mask: Form white;
        combinationRule: Form over.
    CursorManager hand change!

fill
        "Initiate fill mode."
    self newState: #fill:.
    CursorManager hair change!

fill: aCharacter
        "Fill the region that surrounds the point at which the select
         button is released."
    | aChar |
    aChar := aCharacter.
    [aChar == EndSelectFunction]
        whileFalse: [aChar := Terminal read].
    self fillAt: Cursor offset!

initialize: aRectangle
        "Initialize the receiver after the
         window is opened."
    CursorManager execute change.
    lastCursor := CursorManager normal.
    Display fill: aRectangle
        rule: Form over
        mask: Form white.
    self initPen: Display;
        clipRect: aRectangle.
    scanner := CharacterScanner new
        initialize: aRectangle
        font: SysFont.
    nibSize := sourceForm extent.
    nibMask := halftone.
    nibRule := rule := Form over.
    state := #draw:.
    ^Form new
        width: (aRectangle width + 100
            min: 640)
        height: (aRectangle height + 100
            min: 400)
        initialByte: 16rFF!

input: aCharacter
        "A select functions has occured.  Perform the current state."
    self perform: state with: aCharacter!

inputChar: aCharacter
        "A key has been typed, display the character at the
         cursor position."
    | aString |
    aString := ' '.
    aString at: 1 put: aCharacter.
    scanner mask: nibMask.
    scanner
        display: aString
        from: 1
        to: 1
        at: Cursor offset - scanner frame origin.
    Cursor offset:
        Cursor offset +
        ((scanner font charWidth: aCharacter) @ 0)!

line
        "Initiate line drawing."
    self newState: #line:!

line: aCharacter
        "Draw a straight line from the position where the select button
         is pressed to the position at which the select button is
         released."
    | aChar point |
    aChar := aCharacter.
    [aChar == SelectFunction]
        whileFalse: [aChar := Terminal read].
    self defaultNib: 1@1;
        mask: Form white;
        combinationRule: Form reverse;
        place: (point := Cursor offset);
        drawTo: point.
    [Terminal read == EndSelectFunction]
        whileFalse: [
            self drawTo: point.
            self drawTo: (point := Cursor offset)].
    self drawTo: point.
    self reset;
        goto: Cursor offset!

lineFeed
        "The return or line-feed character has been pressed, advance the
         cursor to the next line."
    Cursor offset:
        scanner frame left + 3 @
        (Cursor offset y + scanner font height)!

menu
        "Private - Answer the graphics pane menu."
    ^menu!

newState: aState
        "Make aState the current state."
    state := aState.
    CursorManager normal change.
    self reset!

open
        "Open a free drawing window."
    | topPane |
    menu := Menu
        labels: 'clear\change size\change color\change rule\change font\draw\erase\line\rectangle\circle\ellipse\fill\copy\paste\bit edit' withCrs
        lines: #(1 5 7 12)
        selectors: #(clear changeSize changeMask changeRule changeFont draw erase line rectangle circle ellipse fill copyGraph pasteGraph bitEdit).
    topPane := TopPane new
        model: self;
        label: 'Form Editor';
        minimumSize: Display extent * 2 // 3;
        yourself.
    topPane addSubpane:
        (pane :=
         FreeDrawPane new
            model: self;
            name: #initialize:;
            menu: #menu;
            change: #change).
    topPane dispatcher open scheduleWindow!

pasteGraph
        "Prompt the user for a name of a form in the GraphDictionary
         and paste the contents of the form at the current cursor
         position."
    | aPoint newForm names aName |
    names := PictureDictionary keys asSortedCollection asArray.
    names size = 0
        ifTrue: [^nil].
    aName := (Menu
        labelArray: names
        lines: Array new
        selectors: names)
            popUpAt: Cursor offset.
    aName isNil ifTrue: [^nil].
    newForm := PictureDictionary at: aName.
    "Prompt for a location"
    aPoint := PointDispatcher pointFromUserDisplaying:
        (Cursor offset extent: newForm extent).
    sourceForm := newForm.
    self destRect: (aPoint extent: newForm extent);
        mask: nil;
        combinationRule: nibRule;
        copyBits.
    self reset!

rectangle
        "Initiate the rectangle drawing state."
    self newState: #rectangle:.
    CursorManager origin change!

rectangle: aCharacter
        "Draw a rectangle whose origin is the position at which the
         select button is pressed and whose corner is where the select
         button is released."
    | aChar origin corner |
    aChar := aCharacter.
    [aChar == SelectFunction]
        whileFalse: [aChar := Terminal read].
    CursorManager corner change.
    self defaultNib: 1@1;
        mask: Form white;
        combinationRule: Form reverse;
        place: (origin := corner := Cursor offset);
        rectangleDraw: origin corner: corner.
    [Terminal read == EndSelectFunction]
        whileFalse: [
            self rectangleDraw: origin corner: corner.
            self rectangleDraw: origin
                corner: (corner := Cursor offset)].
    self rectangleDraw: origin corner: corner;
        reset;
        rectangleDraw: origin
            corner: (Cursor offset).
    CursorManager origin change!

rectangleDraw: origin corner: corner
        "Draw a rectangle with origin and corner."
    self place: origin;
        goto: corner x @ origin y;
        goto: corner;
        goto: origin x @ corner y;
        goto: origin!

rectangleErase: origin corner: corner
        "Draw a white rectangle with origin and corner."
    rule := Form over.
    halftone := Form white.
    self rectangleDraw: origin
        corner: corner.
    self reset!

reframe: aRectangle
        "Reframe the pane."
    self clipRect: aRectangle.
    scanner isNil ifFalse:
        [scanner reframe: aRectangle]!

reset
        "Reset the drawing pen to the nib defaults."
    self defaultNib: nibSize;
        mask: nibMask;
        combinationRule: nibRule!

showWindow
        "Display the part of the window updated by the bit editor."
    bitForm isNil
        ifFalse: [
            bitForm display.
            bitForm := nil]! !
