"This file installs Prolog/V, Smalltalk/V's Prolog environment.  To
 install this application evaluate the following expression (note
 that since there is no '\' character in the path name, the file
 must be in the current directory):

    (File pathName: 'prolog.st') fileIn; close.

 The file prolog.doc contains a brief explanation and user's
 guide.

 To open a Logic Browser, evaluate the following expression:

    LogicBrowser new open

"!
Transcript cr; show: 'Filing in class definitions'!

ClassReader subclass: #LogClassReader
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: '' !

ClassHierarchyBrowser subclass: #LogicBrowser
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: '' !

Collection subclass: #List
  instanceVariableNames:
    'content link '
  classVariableNames: ''
  poolDictionaries: '' !

List subclass: #EmptyList
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: '' !

List subclass: #Relation
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: '' !

Object subclass: #LogicDontCare
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: '' !

Object subclass: #LogicUnbound
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: '' !

Object subclass: #LogicRef
  instanceVariableNames:
    'content '
  classVariableNames: ''
  poolDictionaries: '' !

Object subclass: #Logic
  instanceVariableNames:
    'answer varList unboundVars tempResult '
  classVariableNames:
    'Database BinaryOp '
  poolDictionaries: '' !

Logic subclass: #Prolog
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: '' !

Prolog subclass: #Family
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: '' !
                      
Prolog subclass: #Doctor
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: '' !
                 
Prolog subclass: #Dungeon
  instanceVariableNames:
    'dangerousRooms animator pen '
  classVariableNames: ''
  poolDictionaries: '' !
Transcript cr; show: 'Filing in LogClassReader methods'!
             
!LogClassReader methods !

fileInFrom: aStream
        "Read chunks from aStream until an empty chunk
         (a single '!!') is found.  Compile each chunk
         as a method for the class described by the
         receiver.  Log the source code of
         the method to the change log."
    | source result changeLog |
    changeLog := Sources at: 2.
    changeLog setToEnd.
    self instanceHeaderOn: changeLog.
    [(source := aStream nextChunk) isEmpty]
        whileFalse:[
            result := class compileLogic: source.
            result notNil
                ifTrue: [result value sourceString: source]].
    changeLog
        nextChunkPut: '';
        flush!

instanceHeaderOn: aStream
        "Write a header to aStream which identifies
         the class described by the receiver.  The header
         precedes the source code for the methods (as
         opposed to class methods)."
    aStream
        cr;
        nextPut: $!!;
        nextPutAll: class name;
        space;
        nextPutAll: 'logicMethods !!'! !

Transcript cr; show: 'Filing in Behavior methods'!

!Behavior methods !

compileLogic: codeString
        "Compile the Prolog method contained in codeString.
         The class to use for resolving variables is the receiver.
         If there are no errors, add the method to the receiver
         messageDictionary and answer the Association with the
         message selector as the key and the compiled method
         as the value.  If there is an error, answer nil."
    | answer |
    answer := LCompiler
        compile: codeString
        in: self.
    answer notNil
        ifTrue: [
            self
                addSelector: answer key
                withMethod: answer value].
    ^answer!

compileLogic: codeString notifying: requestor
        "Compile the Prolog method contained in codeString."
    | answer |
    answer := LCompiler
        compile: codeString
        in: self
        notifying: requestor
        ifFail: [^nil].
    self addSelector: answer key withMethod: answer value.
    ^answer!

logicMethods
        "Answer an instance of LogClassReader
         initialized for the receiver."
    ^LogClassReader forClass: self! !
             
Transcript cr; show: 'Filing in LogicBrowser methods'!

!LogicBrowser methods !

accept: codeString from: aDispatcher
        "Private - Accept codeString as an updated
         method or class specification and compile it.
         Notify aDispatcher if the compiler detects errors."
    | result class |
    methodSelectedLast
        ifFalse: [
            ^self acceptClass: codeString from: aDispatcher].
    class := instanceSelectedLast
        ifTrue: [selectedClass]
        ifFalse: [selectedClass class].
    result := self compile: codeString
        notifying: aDispatcher
        in: class.
    result isNil
        ifTrue: [^false]
        ifFalse: [
            Smalltalk
                logPrologSource: codeString
                forSelector: result key
                inClass: class.
            result key == selectedMethod
                ifFalse: [
                    selectedMethod := result key.
                    self
                        changed: #selectors
                        with: #restoreSelected:
                        with: selectedMethod].
            ^true]!

compile: codeString
    notifying: aDispatcher
    in: aClass
    ^aClass compileLogic: codeString
        notifying: aDispatcher!

fileOut
        "Private - Write the source for the selected class
         in chunk file format to a file named with the class
         name reduced to 8 characters, extension 'cls'."
    | file |
    selectedClass isNil
        ifTrue: [^self].
    file := Disk newFile:
        (File
            fileName: selectedClass name
            extension: (String with: $c with: $l with: $s)).
    file lineDelimiter: 10 asCharacter.
    selectedClass fileOutOn: file.
    file nextChunkPut: String new.
    (LogClassReader forClass: selectedClass class) fileOutOn: file.
    (LogClassReader forClass: selectedClass) fileOutOn: file.
    file close!  

open
        "Open a Prolog window."
    self openOn: (Array with: Prolog)!

openOn: classCollection
        "Create a logic hierarchy browser window."
    | topPane oneLineHeight listLineHeight ratio |
    hiddenClasses := Set new.
    instanceSelectedLast := true.
    ratio := 2 / 5.
    self update: classCollection.
    topPane := TopPane new
        label: 'Logic Browser';
        minimumSize: 20 * SysFontWidth
             @ (10 * SysFontHeight);
        yourself.
    topPane addSubpane:
        (ListPane new
            model: self;
            name: #hierarchy;
            change: #hierarchy:;
            menu: #menu;
            framingRatio:
                (0 @ 0 extent: 1/2 @ ratio)).
    topPane addSubpane:
        (ListPane new
            model: self;
            name: #selectors;
            change: #selector:;
            menu: #selectorMenu;
            framingRatio:
                (1/2 @ 0 extent: 1/2 @ ratio)).
    topPane addSubpane:
        (TextPane new
            model: self;
            name: #text;
            change: #accept:from:;
            framingRatio: (0 @ (2/5)
                        corner: 1 @ 1)).
    topPane dispatcher open scheduleWindow!

template
        "Private - Answer the
         template for methods."
    ^'    "comment"
predicate(term1, term2) :-
    goal1(term3),
    goal2(term4).'! !
Transcript cr; show: 'Filing in List methods'!

!List class methods !

head: aObject tail: bObject
    ^self basicNew
        head: aObject;
        tail: bObject;
        yourself!

new
    ^EmptyList new!

with: anObject
    ^self basicNew
        head: anObject;
        tail: EmptyList new!

with: anObject with: bObject
    ^self basicNew
        head: anObject;
        tail: (List with: bObject)! !


!List methods !

add: anObject
    link class == EmptyList
        ifTrue: [
            link := List with: anObject]
        ifFalse: [
            (link isKindOf: List)
                ifTrue: [link add: anObject]
                ifFalse: [self error: 'can''t add to non-list']].
    ^anObject!

addList: aList
    link class == EmptyList
        ifTrue: [
            link := aList]
        ifFalse: [
            (link isKindOf: List)
                ifTrue: [link addList: aList]
                ifFalse: [self error: 'can''t add to non-list']].
    ^aList!

at: anInteger
        "Answer the element at the position anInteger."
    | answer |
    anInteger < 1
        ifTrue: [self error: 'index less than 1'].
    anInteger = 1
        ifTrue: [^content]
        ifFalse: [
            self isLastNode
                ifFalse: [^link at: anInteger - 1]
                ifTrue: [self error: 'out of range']]!

do: aBlock
        "Answer the receiver.  For each element in the receiver,
         evaluate aBlock with that element as the argument."
    | aNode |
    aNode := self.
    [aBlock value: aNode head.
     aNode := aNode tail.
     (aNode isKindOf: List) ifFalse: [^self].
     aNode isEmpty]
        whileFalse: []!

head
    ^content!

head: anObject
    content := anObject!

isEmpty
    ^false!

isLastNode
        "Answer true if receiver's link contains
         an EmptyList."
    ^link class == EmptyList!

printOn: aStream
    aStream nextPut: $[.
    self printOne: aStream.
    aStream nextPut: $]!

printOne: aStream
    content printOn: aStream.
    aStream nextPut: $ .
    (link isKindOf: List)
        ifTrue: [link printOne: aStream]
        ifFalse: [
            aStream nextPutAll: '| '.
            link printOn: aStream]!

size
    ^1 +
        ((link isKindOf: List)
            ifTrue: [link size]
            ifFalse: [0])!

tail
    ^link!

tail: anObject
    link := anObject! !
Transcript cr; show: 'Filing in EmptyList methods'!

!EmptyList class methods !

new
    ^self basicNew! !


!EmptyList methods !

= anObject
    ^anObject class == EmptyList!

add: anObject
    | aList |
    aList := List with: anObject.
    aList become: self.
    ^aList!

do: aBlock
    ^self!

head
        "Error, empty list does not have head."
    ^self error: 'Message not understood: head'!

isEmpty
    ^true!

printOne: aStream
        "Print nothing."!

size
    ^0!

tail
        "Error, empty list does not have tail."
    ^self error: 'Message not understood: tail'! !

Transcript cr; show: 'Filing in Relation methods'!

!Relation class methods !

new
    ^super basicNew! !


!Relation methods !

add: anObject
    (content isNil and: [link isNil])
        ifTrue: [
            content := anObject.
            link := EmptyList new]
        ifFalse: [super add: anObject].
    ^anObject!

printOn: aStream
    content printOn: aStream.
    aStream nextPut: $(.
    (link isKindOf: List)
        ifTrue: [link printOne: aStream]
        ifFalse: [link printOn: aStream].
    aStream nextPut: $)! !
Transcript cr; show: 'Filing in SystemDictionary methods'!

!SystemDictionary methods !

logPrologSource: aString forSelector: aSymbol inClass: aClass
        "Append the source code aString to the
         change log for Prolog method with selector
         aSymbol in aClass.  Make compiled
         method point to source in change log."
    | changeLog |
    changeLog := Sources at: 2.
    changeLog
        setToEnd;
        cr;
        cr;
        nextPut: $!!;
        nextPutAll: aClass name , ' logicMethods !!'.
    (aClass compiledMethodAt: aSymbol) sourceString: aString.
    changeLog
        setToEnd;
        nextChunkPut: String new;
        flush! !
Transcript cr; show: 'Filing in LogicDontCare methods'!

!LogicDontCare class methods ! !


!LogicDontCare methods !

= anObject
    ^true!

printOn: aStream
    aStream nextPutAll: '_';
        nextPut: $ ! !

Transcript cr; show: 'Filing in LogicUnbound methods'!

!LogicUnbound methods !

printOn: aStream
    'Unbound' printOn: aStream! !
Transcript cr; show: 'Filing in LogicRef methods'!

!LogicRef class methods !

new
    ^super new value: LogicUnbound new! !


!LogicRef methods !

= anObject
    ^self value = anObject!

== anObject
    ^self value == anObject!

at: anObject
    ^self value at: anObject!

content
    ^content!

content: anObject
    content := anObject!

doesNotUnderstand: aMessage
    ^self value
        perform: aMessage selector
        withArguments: aMessage arguments!

lastNode
    | aNode |
    aNode := self.
    [aNode content class == LogicRef]
        whileTrue: [aNode := aNode content].
    ^aNode!

printOn: aStream
    self value printOn: aStream!

value
    | aNode |
    aNode := content.
    [aNode class == LogicRef]
        whileTrue: [aNode := aNode content].
    ^aNode!

value: anObject
    | node |
    node := self lastNode.
    anObject class == LogicRef
        ifTrue: [
            node == anObject lastNode
                ifFalse: [node content: anObject]]
        ifFalse: [
            node content: anObject]! !
Transcript cr; show: 'Filing in Logic methods'!

!Logic class methods !

database
    ^Database!

initialize
        "Initialize Database."
    Database := Dictionary new!

new
    ^super new initialize! !


!Logic methods !

allValue: aLogicRef
        "Copy all the values to an instance
         the same class as the content of aLogicRef.
         If aLogicRef's content is a List, walk
         through all elements."
    | answer aList |
    (aLogicRef isKindOf: LogicRef)
        ifTrue: [answer := aLogicRef value]
        ifFalse: [answer := aLogicRef].
    (answer isKindOf: List)
        ifTrue: [
            answer isEmpty ifTrue: [^answer].
            aList := answer.
            answer := answer class with:
                (self allValue: aList head).
            answer tail: (self allValue: aList tail)].
    ^answer!

buildQuery: aList
        "Private"
    | pred block |
    pred := aList head.
    block :=
        (aList tail isEmpty
            ifTrue: [[self inferAnswer]]
            ifFalse: [self buildQuery: aList tail]).
    ^[self perform: pred head with:
        (Association
            key: pred tail
            value: block)]!

doesNotUnderstand: aMessage
    | temp |
    temp := aMessage selector.
    temp := (temp copyFrom: 1 to: temp size - 1)
        asSymbol.
    temp := Database at: temp ifAbsent: [
        ^self error: 'Message not understood: ', aMessage selector
        ].
    ^self retrieve: (aMessage arguments at: 1)
        from: temp!

doGoal: aRelation continue: aBlock
    "Try to satisfy aRelation as a goal."
    | blk mark |
    mark := unboundVars size.
    ((aRelation isKindOf: List) and: [
    aRelation head class == Symbol])
        ifFalse: [self error: 'not a goal'].
    blk := self
        perform: (aRelation head, ':') asSymbol
        with: (Association
            key: aRelation tail
            value: aBlock).
    [blk class == Context]
        whileTrue: [blk := blk value].
    self unbindVar: mark!

doOneGoal: aRelation
    "Try to satisfy aRelation as a goal.
     Answer true if success, else false."
    | blk mark |
    mark := unboundVars size.
    ((aRelation isKindOf: List) and: [
    aRelation head class == Symbol])
        ifFalse: [self error: 'not a goal'].
    blk := self
        perform: (aRelation head, ':') asSymbol
        with: (Association
            key: aRelation tail
            value: [^true]).
    [blk class == Context]
        whileTrue: [blk := blk value].
    self unbindVar: mark.
    ^false!

first: assoc
        "Private -- Answer the first argument."
    | answer |
    answer := assoc key head.
    ^self value: answer!

inferAnswer
        "Private -- Add an Array of values to
         the answer collection."
    | anArray i |
    anArray := Array new: varList size.
    i := 1.
    varList do: [:x |
        anArray at: i put: (self allValue: x).
        i := i + 1].
    answer add: anArray!

inferSmalltalk: aList var: bList
        "Private -- Handle query from Smalltalk. aList
         contains a list of predicates and
         bList is a list of unbound variables."
    | blk |
    varList := bList.
    answer := OrderedCollection new.
    blk := self buildQuery: aList.
    [(blk := blk value) class == Context]
        whileTrue: [].
    self unbindVar: 0.
    answer size = 0
        ifTrue: [^nil]
        ifFalse: [^answer asArray]!

initialize
        "Private -- Initialize instance variables."
    unboundVars := OrderedCollection new: 2000!

match: x with: y
        "Private -- Answer true if x matches y, else
         answer false."
    | xLast yLast |
 "Transcript nextPutAll: x printString;nextPut: $ ;
 nextPutAll: y printString;cr."
    (x class == LogicDontCare or: [y class == LogicDontCare])
         ifTrue: [^true].
    (x class == LogicRef)
        ifTrue: [
            xLast := x lastNode.
            xLast value class == LogicUnbound
                ifTrue: [
                    unboundVars add: xLast.
                    xLast value: y.
                    ^true]
                ifFalse: [xLast := xLast value]]
        ifFalse: [xLast := x].
    (y class == LogicRef)
        ifTrue: [
            yLast := y lastNode.
            yLast value class == LogicUnbound
                ifTrue: [
                    unboundVars add: yLast.
                    yLast value: x.
                    ^true]
                ifFalse: [
                    yLast := yLast value]]
        ifFalse: [yLast := y].
    (xLast class == List) ifTrue: [
        (yLast class == List)
            ifTrue: [
                (self match: xLast head with: yLast head)
                    ifTrue: [^self
                        match: xLast tail
                        with: yLast tail]]
            ifFalse: [^false]].
    (xLast class == Relation) ifTrue: [
        (yLast class == Relation)
            ifTrue: [
                (self match: xLast head with: yLast head)
                    ifTrue: [^self
                        match: xLast tail
                        with: yLast tail]]
            ifFalse: [^false]].
    ^xLast = yLast!

retrieve: assoc from: dataList
        "Retrieve a fact in assoc from
         Database entry dataList."
    | aList |
    dataList isEmpty ifTrue: [^self].
    aList := dataList.
    [aList isLastNode] whileFalse: [
        self unify: assoc key
            with: aList head
            then: assoc value.
        aList := aList tail].
    self unify: assoc key
        with: aList head
        then: [^assoc value]!

second: assoc
        "Private -- Answer the second argument."
    | answer |
    answer := assoc key tail head.
    ^self value: answer!

sturctureHead: aSymbol arguments: anInteger
        "Answer a structure whose functor is
         aSymbol and arguments are anInteger
         number of DontCares."
    | aRelation |
    aRelation := Relation with: aSymbol.
    anInteger timesRepeat: [
        aRelation add: LogicDontCare new].
    ^aRelation!

unbindVar: mark
        "Private -- Unbind previously bound variables
         down to mark."
    | index |
    [(index := unboundVars size) > mark]
        whileTrue: [
            (unboundVars at: index)
                content: LogicUnbound new.
            unboundVars removeLast]!

unify: x with: y then: aBlock
        "Private -- If x unifies with y, then execute
         aBlock. Answer self."
    | mark |
" Transcript nextPutAll: x printString;nextPut: $ ;
 nextPutAll: y printString;cr."
    mark := unboundVars size.
    (self match: x with: y)
        ifTrue:
            [tempResult := aBlock.
             [(tempResult := tempResult value).
              tempResult class == Context]
                whileTrue: []].
    self unbindVar: mark!

value: aRef
        "Private -- Answer the value of the
         logic ref, aRef."
    aRef class == LogicRef
        ifTrue: [^aRef value]
        ifFalse: [^aRef]! !
Transcript cr; show: 'Filing in Prolog methods'!
                             
!Prolog methods !

and: assoc
    "Try to satisfy a list of goals."
    | aList |
    aList := assoc key.
    [aList isEmpty] whileFalse: [
        (self doOneGoal: aList head)
            ifTrue: [aList := aList tail]
            ifFalse: [^self]].
    ^assoc value! !

!Prolog logicMethods !

"Bind x to the n'th argument of structure."
arg(n, structure, x) :-
    integer(n),
    is(true, structure value isKindOf: List),
    is(x, (self allValue: structure)
        at: n + 1).! !

!Prolog methods !

asserta: assoc
        "Add the fact in assoc to the
         beginning of the database."
    | arg aList |
    assoc key size = 1 ifFalse: [^self].
    arg := self allValue: (self first: assoc).
    aList := Database at: arg head ifAbsent: [^self].
    Database at: arg head put:
        (List head: arg tail tail: aList).
    ^assoc value!

assertz: assoc
        "Add the fact in assoc to the
         end of the database."
    | arg aList |
    assoc key size = 1 ifFalse: [^self].
    arg := self allValue: (self first: assoc).
    aList := Database at: arg head ifAbsent: [^self].
    aList add: arg tail.
    ^assoc value!

atom: assoc
        "Succeed if the argument is a Symbol."
    assoc key size = 1 ifFalse: [^self].
    (self first: assoc) class == Symbol
        ifTrue: [^assoc value]!

atomic: assoc
        "Succeed if the argument is a Symbol
         or an integer."
    | arg |
    assoc key size = 1 ifFalse: [^self].
    ((arg := self first: assoc) class == Symbol
    or: [arg isKindOf: Integer])
        ifTrue: [^assoc value]!

call: assoc
        "Execute the argument as a predicate."
    | blk |
    assoc key size = 1 ifFalse: [^self].
    blk := self first: assoc.
    ((blk isKindOf: List) and: [
    blk head class == Symbol])
        ifFalse: [self error: 'not a goal'].
    ^[self perform: (blk head, ':') asSymbol
            with: (Association key: blk tail
                value: assoc value)]!

consult: assoc
        "Ask the second argument as the receiver
         (a Smalltalk expression) to solve the goals
         contained in the first argument, e.g.
            consult(father(x, 'John'), family new)."
    | blk |
    assoc key size = 2 ifFalse: [^self].
    blk := self first: assoc.
    blk class == Relation
        ifFalse: [self error: 'first argument must be predicate'].
    ^[(self second: assoc) value
        perform: (blk head, ':') asSymbol
        with: (Association key: blk tail
               value: assoc value)]!

cut: assoc
        "Succeed always, exit the containing method
         during backtacking."
    | blk mark |
    assoc key size = 1 ifFalse: [^self].
    mark := unboundVars size.
    blk := assoc value.
    [(blk := blk value) class == Context]
        whileTrue: [].
    self unbindVar: mark.
    assoc key head value!

database: assoc
        "Create a database which can hold facts
         being asserted or extracted. The argument
         must be a structure."
    | arg functor source |
    assoc key size = 1 ifFalse: [^self].
    arg := self first: assoc.
    ((arg isKindOf: Relation) and:
    [arg head isKindOf: Symbol])
        ifFalse: [self error: 'argument is not a fact'].
    functor := (arg head, ':') asSymbol.
    Database at: arg head ifAbsent: [
        Database at: arg head put: List new].
    ^assoc value!

eq: assoc
        "Succeed if the two arguments can
         be unified."
    assoc key size = 2 ifFalse: [^self].
    self unify: assoc key head
        with: assoc key tail head
        then: [^assoc value]!

equiv: assoc
        "Succeed if the two arguments
         are the same object."
    assoc key size = 2 ifFalse: [^self].
    (self first: assoc) == (self second: assoc)
        ifTrue: [^assoc value]!

fail: assoc
        "Fail always."! !

!Prolog logicMethods !

"The first argument is unbound."
functor(structure, head, arguments) :-
    var(structure),
    atom(head),
    integer(arguments),
    is(structure, self
        sturctureHead: head
        arguments: arguments).
    "The first argument can be either a
     Relation or a List whose first
     element is a Symbol."
functor([sHead | sTail], head, arguments) :-
    atom(sHead),
    eq(head, sHead),
    is(arguments, sTail value size).! !

!Prolog methods !

ge: assoc
        "Succeed if the first argument is
         greater than or equal to the second."
    assoc key size = 2 ifFalse: [^self].
    (self first: assoc) >= (self second: assoc)
        ifTrue: [^assoc value]!

gt: assoc
        "Succeed if the first argument is
         greater than the second."
    assoc key size = 2 ifFalse: [^self].
    (self first: assoc) > (self second: assoc)
        ifTrue: [^assoc value]!

halt: assoc
        "Halt the execution.  Smalltalk debugger
         can be invoked at this point."
    assoc key size = 0 ifFalse: [^self].
    self halt!

integer: assoc
        "Succeed if the argument is an Integer."
    assoc key size = 1 ifFalse: [^self].
    ((self first: assoc) isKindOf: Integer)
        ifTrue: [^assoc value]!

is: assoc
        "Execute the second argument as a Smalltalk
         expression and then unify its result with
         the first argument."
    assoc key size = 2 ifFalse: [^self].
    self unify: assoc key head
        with: (self value: assoc key tail head) value
        then: [^assoc value]!

le: assoc
        "Succeed if the first argument is
         less than or equal to the second."
    assoc key size = 2 ifFalse: [^self].
    (self first: assoc) <= (self second: assoc)
        ifTrue: [^assoc value]!

lt: assoc
        "Succeed if the first argument is
         less than the second."
    assoc key size = 2 ifFalse: [^self].
    (self first: assoc) < (self second: assoc)
        ifTrue: [^assoc value]!

ne: assoc
        "Succeed if the first argument is
         greater than the second."
    assoc key size = 2 ifFalse: [^self].
    (self first: assoc) ~= (self second: assoc)
        ifTrue: [^assoc value]!

nl: assoc
        "Write an end of line to
         Transcript."
    assoc key size = 0 ifFalse: [^self].
    Transcript cr.
    ^assoc value!

nonvar: assoc
        "Succeed if the argument is not an
         unbound variable."
    assoc key size = 1 ifFalse: [^self].
    (self first: assoc) class == LogicUnbound
        ifFalse: [^assoc value]!

not: assoc
        "Succeed if the evaluation of the argument
         as a goal fails."
    | blk |
    assoc key size = 1 ifFalse: [^self].
    blk := self first: assoc.
    (blk isKindOf: List)
        ifFalse: [self error: 'argument must be goal'].
    blk :=
        self perform: (blk head , ':') asSymbol
            with:
                (Association key: blk tail
                    value: [^self]).
    [blk class == Context] "value until real success"
        whileTrue: [blk := blk value].
    ^assoc value!

notEquiv: assoc
        "Succeed if the two arguments
         are not the same object."
    assoc key size = 2 ifFalse: [^self].
    (self first: assoc) == (self second: assoc)
        ifFalse: [^assoc value]!

or: assoc
    "Disjunction predicate."
    assoc key size = 2 ifFalse: [^self].
    self doGoal: (self first: assoc) continue: assoc value.
    self doGoal: (self second: assoc) continue: assoc value!

purgeDatabase: assoc
        "Purge a database identified by the
         structure in the argument."
    | arg |
    assoc key size = 1 ifFalse: [^self].
    arg := self first: assoc.
    ((arg isKindOf: Relation) and:
    [arg head isKindOf: Symbol])
        ifFalse: [self error: 'argument is not a fact'].
    Database at: arg head ifAbsent: [^self].
    Database removeKey: arg head.
    ^assoc value!

read: assoc
        "Read a Smalltalk expression, evaluate it and
         answer the result.  This can be used to input
         a Symbol, a number, a quoted string, etc."
    | answer |
    assoc key size = 1 ifFalse: [^self].
    answer := Prompter prompt: 'type a line and return'
        defaultExpression: ''.
    self unify: assoc key head
        with: answer
        then: [^assoc value]!

repeat: assoc
        "Succeed always, even
         during backtacking."
    | blk mark |
    assoc key size = 0 ifFalse: [^self].
    mark := unboundVars size.
    [true] whileTrue: [
        blk := assoc value.
        [(blk := blk value) class == Context]
            whileTrue: [].
        self unbindVar: mark]!

retract: assoc
        "Delete the first fact unified with the
         one in assoc from the database. Delete
         the next one during backtrack."
    | arg aList bList |
    assoc key size = 1 ifFalse: [^self].
    arg := self first: assoc.
    [aList := Database at: arg head ifAbsent: [^self].
     aList isEmpty ifTrue: [^self].
     self unify: arg tail
        with: aList head
        then: [Database at: arg head put: aList tail.
               assoc value value].
     aList == (Database at: arg head)]
        whileFalse: [].
    [(bList := aList tail) isEmpty] whileFalse: [
        self unify: arg tail
            with: aList tail head
            then: [aList tail: aList tail tail.
                   bList := aList.
                   assoc value value].
        aList := bList.
        aList isEmpty ifTrue: [^self]]!

runtime: assoc
        "Unify the first argument with
         the system time in milliseconds."
    assoc key size = 1 ifFalse: [^self].
    self unify: assoc key head
        with: Time millisecondClockValue
        then: [^assoc value]!

true: assoc
        "Succeed always."
    assoc key size = 0 ifFalse: [^self].
    ^assoc value! !

!Prolog logicMethods !

"Convert a structure to a list (=..)."
univ(structure, list) :-
    nonvar(structure),
    is(true, structure value class == Relation),
    is(x, self allValue: structure),
    is(list, List
        head: x value head
        tail: x value tail).
univ(structure, [head | list]) :-
    atom(head),
    is(x, self allValue: list),
    is(structure, Relation
        head: head value
        tail: x value).! !

!Prolog methods !

var: assoc
        "Succeed if the argument is an
         unbound variable."
    assoc key size = 1 ifFalse: [^self].
    (self first: assoc) class == LogicUnbound
        ifTrue: [^assoc value]!

write: assoc
        "Write all the arguments to Transcript."
    | value |
    assoc key do: [:argument |
        value := self allValue: argument.
        Transcript nextPutAll:
            ((value isKindOf: String)
                ifTrue: [value]
                ifFalse: [value printString])].
    ^assoc value! !
Transcript cr; show: 'Filing in Family methods'!

!Family logicMethods !

father('John', 'Mary').  "John is the father of Mary"
father('John', 'David').
father('David', 'Jack').
father('Arthur', 'Nancy').!

"x is the grandPa of y."
grandPa(x, y) :-
    father(x, z),
    or(father(z, y),
       mother(z, y)).!  
 
"Return the health history of x."
health(x, history) :-
    consult( healthHistory(x, history), Doctor new).!

"Mary is the mother of Nancy."
mother('Mary', 'Nancy').! !
Transcript cr; show: 'Filing in Collection methods'!

!Collection methods !

asList
    | aList pointer |
    aList := EmptyList new.
    self do: [:i |
        aList add: i.
        pointer isNil
            ifTrue: [pointer := aList]
            ifFalse: [pointer := pointer tail]].
    ^aList! !

"evaluate"
Logic initialize!
Transcript cr; show: 'Filing in Dungeon methods'!
                   
!Dungeon logicMethods !

"Adjust the y coordinate by Aspect ratio."
adjust(result, y) :-
    is(result, (y value * Aspect) truncated).!

"Fail if room is one of roomsToAvoid and
 go back to previousRoom. Else succeed."
avoid(room,roomsToAvoid,previousRoom) :-
    member(room,roomsToAvoid),
    goto(previousRoom), !!, fail().
avoid(_,_,_).!

"Draw a map of the dungeon."
draw() :- drawArc(), fail().
draw() :- drawNode(), fail().
draw().!

"Draw the path between two rooms."
drawArc() :-
    gallery(x, y),
    position(x, a, b),
    position(y, c, d),
    is(_, pen
            defaultNib: 3 @ 2;
            place: a value @ b value;
            goto: c value @ d value).!

"Call a Smalltalk method to draw a room."
drawNode() :-
    position(x, a, b),
    is(_,
        Display white:
            (a value - 35 @ (b value - 15)
            extent: 70 @ 30)),
    is(_, pen
            place: a value @ b value;
            centerText: x value font: SysFont).!

"Define database of all connecting rooms."
gallery(#entry,#monsters).
gallery(#entry,#stadium).
gallery(#stadium,#hell).
gallery(#stadium,#dolphin).
gallery(#stadium,#pond).
gallery(#exit,#gold).
gallery(#robbers,#gold).
gallery(#stadium,#robbers).
gallery(#pond,#gold).
gallery(#dolphin,#exit).
gallery(#monsters,#gold).!

"Move from a room here to a room there."
go() :-
    initialize(),
    draw(),
    is(_, animator setBackground), 
    route(#entry,#exit,[#entry]).!

"Move the animal to nextRoom."
goto(nextRoom) :-
    position(nextRoom, x, y),
    is(_, animator
        tell: 'Snoopy'
        goto: x value - 30 @ (y value - 25)).!

"Check if adventure is succeeded."
isFinished(visitedRooms) :-
    member(#gold,visitedRooms),
    end().
isFinished([_, x | _]) :-
    goto(x), !!, fail().!

"Succeed if the first argument is a
 member of a List in the second argument."
member(x,[x|_]).
member(x,[_|h]) :- member (x,h).!

"Unify the next neighbor room of x with y."
neighborroom(x,y,visitedRooms) :-
    or(gallery(x,y), gallery(y,x)).
neighborroom(_,_,[_ , x | _]) :-
    goto(x), !!, fail().!

"Define the position of each room on screen."
position(#entry,40,y) :- adjust(y, 80).
position(#stadium,230,y) :- adjust(y, 50).
position(#monsters,40,y) :- adjust(y, 200).
position(#hell,90,y) :- adjust(y, 310).
position(#pond,330,y) :- adjust(y, 160).
position(#gold,460,y) :- adjust(y, 200).
position(#dolphin,380,y) :- adjust(y, 70).
position(#robbers,230,y) :- adjust(y, 310).
position(#exit,580,y) :- adjust(y, 180).!

"Find the route to the next room avoiding
 dangerous or visited rooms."
route(#exit,#exit,visitedRooms) :-
    isFinished(visitedRooms).
route(room,wayOut,visitedRooms) :-
    not(eq(room, #exit)),
    neighborroom(room,nextRoom,visitedRooms),
    not(member(nextRoom,visitedRooms)),
    goto(nextRoom),
    avoid(nextRoom,dangerousRooms,room),
    route(nextRoom,wayOut,[nextRoom|visitedRooms]).! !

!Dungeon methods !

end: arg
        "End the whole thing.  This method
         does not return.  It restarts
         the Scheduler."
    animator tell: 'Snoopy' direction: 90.
    14 timesRepeat: [ "jump 7 times"
            animator tell: 'Snoopy' turn: 180.
            animator tell: 'Snoopy' go: 20].
    animator := nil.
    Menu message:
        'You found gold and got to the exit!!'.
    Scheduler resume!

initialize: arg
        "This is a Smalltalk method, but can be called as
         a Prolog/V predicate.  It initializes instance
         variables and gray the screen.  The animal moving
         speed is also initialized here."
    dangerousRooms := List with: #monsters with: #robbers.
    Display gray.
    animator := Animation new
        initialize: Display boundingBox.
    animator
        add: (Array
            with: (GraphDictionary at: 'dog1')
            with: (GraphDictionary at: 'dog2')
            with: (GraphDictionary at: 'dog3'))
        name: 'Snoopy'
        color: #black;
        speed: 15.
    pen := Pen new.
    ^arg value! !   
Transcript cr; show: 'Filing in Doctor methods'!
                      
!Doctor logicMethods !

"Health history of individuals."
healthHistory('John', '
    John had two times of heart attacks').
healthHistory('Mary', '
    Mary had no serious problems').! !

Transcript cr; show: 'Prolog installed.'!

