"This file installs the Tree Browser application.  To install
 this application evaluate the following expression (note that since
 there is no '\' character in the path name, the file must be in the
 current directory):

    (File pathName: 'tree.st') fileIn; close.

 The file tree.doc contains a brief explanation and user's
 guide.

 "!

(Disk file: 'formedtr.cls') fileIn; close. !
(Disk file: 'fredrwpn.cls') fileIn; close. !

Object subclass: #TreeBrowser
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: '' !

TreeBrowser subclass: #Tree
  instanceVariableNames:
    'pen scanner getSonBlock getNameBlock branchLength form fontHeight fontWidth maxWidth maxHeight firstNode pane '
  classVariableNames: ''
  poolDictionaries: '' !

Object subclass: #TreeBrowser
  instanceVariableNames: ''
  classVariableNames: ''
  poolDictionaries: '' !

TreeBrowser subclass: #Node
  instanceVariableNames:
    'value column row son brother '
  classVariableNames: ''
  poolDictionaries: '' !

!DiskBrowser methods !

createDirectory
        "Private - Request the name of a new directory with
         a Prompter and create a subdirectory with the entered
         name under the selected directory."
    | aDirectory index size aForm aDispatcher |
    aDirectory := self promptForPathName.
    aDirectory isNil
        ifTrue: [^nil].
    CursorManager execute change.
    aDirectory create.
    self buildDirectoryList.
    self selectDirectory: aDirectory.
    self updateTree.
   CursorManager normal change !

directoryListMenu
        "Private - Answer the directory pane menu."
    ^Menu
        labels: 'remove\update\create\tree' withCrs
        lines: Array new
        selectors: #(removeDirectory update createDirectory tree)!

input: aChar
    | dir |
    aChar = EndSelectFunction
        ifTrue: [
            dir := dirTree select: Cursor offset.
            dir isNil ifFalse: [
                Scheduler add:
                    contentsPane topPane dispatcher. 
                contentsPane topPane zapBackup.
                self selectDirectory:
                    (dir := Directory new drive: device;
                        pathName: (dir at: 1)).
                Scheduler run]]!

inputChar: aChar
        "Do nothing when type in tree pane."!

release
    | aTree |
    dirTree isNil
        ifTrue: [super release]
        ifFalse: [
            aTree := dirTree.
            dirTree := nil.
            contentsPane topPane dispatcher
            == Scheduler topDispatcher ifTrue: [
                super release.
                aTree pane topPane
                    dispatcher closeWindow.
                TopPane initWindowClip]]!

removeDirectory
        "Private - Remove the selected directory."
    | dir |
    selectedDirectory isNil
        ifTrue: [^nil].
    CursorManager execute change.
    (Directory new drive: device; pathName: '\') makeCurrent.
    selectedDirectory remove.
    contentsPane name: #directory.
    self
        buildDirectoryList;
        changed: #label;
        changed: #directories
            with: #restore;
        changed: #files;
        changed: #directorySort
            with: #restoreSelected:
            with: 1;
        changed: #directory.
    self updateTree.
    CursorManager normal change!

tree
    | backSlash |
    backSlash := '\'.
    dirTree := (Smalltalk at: #Tree ifAbsent:
        [self error: 'Tree class is missing'])
             new.
    dirTree openOn: (Array with:
            (Array with: backSlash with: backSlash))
            withName: [ :a | a at: 2]
            son: [ :a |
                (Directory new drive: device;
                    pathName: (a at: 1))
                subdirectories asSortedCollection:
                    [:b :c | (b at: 2) < (c at: 2)]]
            branchLength: 32
            font: Font eightLine
            label: 'Directory of ', (String with: device with: $:)
            model: self!

updateTree
    | aForm |
    dirTree isNil ifFalse: [
        aForm := dirTree buildTree: (Array with:
            (Array with: '\' with: '\'))
            font: Font eightLine.
        Scheduler add:
            dirTree pane topPane dispatcher.
        dirTree pane form: aForm;
            update.
        Scheduler run]! !
                             
!Node class methods !

value: anObject column: colInteger row: rowInteger
    "Build a node."
    ^self new
        value: anObject
        column: colInteger
        row: rowInteger
        son: nil!

value: anObject column: colInteger row: rowInteger son: aNode
    "Build a node."
    ^self new
        value: anObject
        column: colInteger
        row: rowInteger
        son: aNode! !


!Node methods !

brother
        "Answer the brother node of the
         receiver."
    ^brother!

brother: aNode
        "Set the brother of the receiver
         to aNode."
    brother := aNode!

column
        "Answer the printing column
         of the receiver node."
    ^column!

contains: aPoint extent: xPoint
        "Answer true if the receiver node
         contains aPoint. Else answer false.
         The extent of the node is xPoint."
    ^(column @ row extent: xPoint)
        containsPoint: aPoint!

row
        "Answer the printing row of
         the receiver node."
    ^row!

son
        "Answer the son node of the receiver."
    ^son!

value
        "Answer the receiver node value
         which can be any object."
    ^value!

value: anObject column: colInteger row: rowInteger son: aNode
        "Initialize a node with anObject as its value,
         colInteger and rowInteger as its printing
         position, and aNode as its son node."
    value := anObject.
    column := colInteger.
    row := rowInteger.
    son := aNode.
    brother := nil! !

!Tree methods !

buildTree: topNodes font: aFont
        "Build a tree out of topNodes and
         print it on form using aFont as
         the printing font. Answer the form."
    fontHeight := aFont height.
    fontWidth := aFont width.
    maxWidth := maxHeight := 0.
    firstNode := (self getTree: topNodes column: 8 row: 0)
        at: 1.                                  "
Transcript nextPutAll: maxWidth printString;
nextPut:$@; nextPutAll: maxHeight printString; cr.  "
    form := Form new
        width: maxWidth + 8
        height: maxHeight
        initialByte: 16rFF.
    pen := Pen new: form.
    scanner := CharacterScanner new
        initialize: (0@0 extent: form extent)
        font: aFont
        dest: form.
    self
        display: firstNode branchStart: nil.
    pen := scanner := nil.
    ^form!

display: aNode branchStart: aPoint
        "Display the value of aNode, its
         bothers and sons at their printing positions.
         Branches are drawn from aPoint (the end of
         the father node) to aNode and its brothers."
    | node |
    node := aNode.
    [node isNil] whileFalse: [
        "print the node"
        scanner
            display: (getNameBlock value: node value)
            at: node column @ node row.
        aPoint isNil ifFalse: [
            pen drawFrom: aPoint to: node column - 2
                @ (fontHeight // 2 + node row)].
        "print sons"
        self display: node son
            branchStart: scanner destX + 1
                @ (fontHeight // 2 + node row).
        node := node brother]!

getTree: values column: col row: rowNumber
        "Build a tree with values stuffed in its top
         nodes. The sons of the tree are derived from
         these values. The top left corner of the rectangle
         containing the tree is col @ rowNumber.
         Answer a two-element array whose first element
         is the first node made out of values, and
         the second is the bottom row of the last node."
    | bigBrother curNode sons row firstSon firstNode |
    row := rowNumber.
    bigBrother := nil.
    firstNode := nil.
    values do: [ :aValue |
        sons := getSonBlock value: aValue.
        sons size = 0
            ifTrue: [ "terminal node"
                curNode :=
                    Node value: aValue
                        column: col
                        row: row.
                row := row + fontHeight + 2.
                maxWidth := maxWidth max:
                            (getNameBlock value: aValue) size
                            * fontWidth + col
                ]
            ifFalse: [ "non-terminal node"
                firstSon :=
                    self getTree: sons
                        column:
                            (getNameBlock value: aValue) size
                            * fontWidth + col + branchLength
                        row: row.
                curNode :=
                    Node value: aValue
                        column: col
                        row: (firstSon at: 1) row +
                            (firstSon at: 2) // 2
                        son: (firstSon at: 1).
                row := maxHeight
                ].
        bigBrother isNil ifFalse: [
            bigBrother brother: curNode].
        bigBrother := curNode.
        firstNode isNil ifTrue: [firstNode := curNode].
    (maxHeight // 8 * maxWidth) > 31000
        ifTrue: [
            ^Array with: firstNode with: curNode row].
        ].
    maxHeight := row.
    ^Array with: firstNode with: curNode row!

input: aChar
        "Do nothing when a character is
         input from the tree pane."!

node: curNode contains: aPoint
        "Answer the value of the node
         containing aPoint starting
         from curNode and walk the tree."
    | value |
    (curNode contains: aPoint
        extent: (getNameBlock value: curNode value) size
            * fontWidth @ fontHeight)
        ifTrue: [^curNode value]
        ifFalse: [
            curNode son isNil
                ifFalse: [value :=
                    self node: curNode son contains: aPoint].
            value isNil
                ifTrue: [
                    curNode brother isNil
                        ifTrue: [^nil]
                        ifFalse: [
                            ^self node: curNode brother
                                contains: aPoint]]
                ifFalse: [^value]]!

openOn: aCollection
    withName: nBlock
    son: sBlock
    branchLength: lenInteger
    font: aFont
    label: aString
    model: aModel
        "Open a tree browser for aCollection of
         topNodes. The son and name of a node can
         be obtained by nBlock and sBlock.
         The horizontal branch length is lenInteger.
         The window label is aString. The controlling
         application is aModel. And the tree will
         be printed using aFont as the font."
    | topPane |
    CursorManager execute change.
    getNameBlock := nBlock.
    getSonBlock := sBlock.
    branchLength := lenInteger.
    topPane := TopPane new
        label: aString;
        minimumSize: 100@100;
        yourself.
    topPane addSubpane:
        (pane := FreeDrawPane new
            model: aModel;
            form: (self buildTree: aCollection font: aFont);
            framingBlock: [:aRect| aRect]).
    CursorManager normal change.
    topPane dispatcher open scheduleWindow!

pane
        "Answer the pane of the receiver tree."
    ^pane!

select: aPoint
        "Answer the value of the node
         that contains aPoint."
    ^self node: firstNode contains:
        aPoint - pane frame origin + pane topCorner! !


